"""
Amatino API Python Bindings
Denomination Module
Author: hugh@amatino.io
"""
from amatino.internal.immutable import Immutable


class Denomination:
    """
    Abstract class defining an interface for units of account. Adopted by
    Custom Units and Global Units.
    """
    def __init__(
        self,
        code: str,
        id_: int,
        name: str,
        priority: int,
        description: str,
        exponent: int
    ) -> None:

        self._code = code
        self._id = id_
        self._name = name
        self._priority = priority
        self._description = description
        self._exponent = exponent
        return

    code = Immutable(lambda s: s._code)
    id_ = Immutable(lambda s: s._id)
    name = Immutable(lambda s: s._name)
    priority = Immutable(lambda s: s._priority)
    description = Immutable(lambda s: s._description)
    exponent = Immutable(lambda s: s._exponent)



"""
Amatino API Python Bindings
AM Time Package
Author: hugh@amatino.io

This module is intended to be private, used indirectly
by public classes, and should not be used directly.
"""
import datetime
from amatino.internal.immutable import Immutable
from amatino.internal.encodable import Encodable
from amatino.api_error import ApiError
from typing import Type
from typing import TypeVar

T = TypeVar('T', bound='AmatinoTime')


class AmatinoTime(Encodable):
    """
    Private - Not intended to be used directly.

    An Amatino-specific time instance, used to convert datetime objects into
    strings of the format expected by the Amatino API
    """
    _FORMAT_STRING = '%Y-%m-%d_%H:%M:%S.%f'

    def __init__(self, date_time: datetime.datetime) -> None:
        if not isinstance(date_time, datetime.datetime):
            raise TypeError('Unexpected datetime type, `datetime` required')
        self._raw_time = date_time.replace(tzinfo=AmatinoTime.UTC())
        return

    string = Immutable(lambda s: s.serialise())
    raw = Immutable(lambda s: s._raw_time)

    def serialise(self) -> str:
        return datetime.datetime.strftime(self._raw_time, self._FORMAT_STRING)

    class UTC(datetime.tzinfo):
        """UTC timezone."""

        def utcoffset(self, dt):
            return datetime.timedelta(0)

        def tzname(self, dt):
            return 'UTC'

        def dst(self, dt):
            return datetime.timedelta(0)

    @classmethod
    def decode(cls: Type[T], data: str) -> T:
        """Return an AmatinoTime instance decoded from a string"""

        if not isinstance(data, str):
            raise ApiError('Unexepected type when decoding AmatinoTime')

        date_time = datetime.datetime.strptime(data, AmatinoTime._FORMAT_STRING)
        amatino_time = cls(date_time)
        return amatino_time



"""
Amatino API Python Bindings
Unexpected Response Type Module
Author: hugh@amatino.io
"""
from amatino.api_error import ApiError
from typing import Any


class UnexpectedResponseType(ApiError):
    _PREFIX = 'Unexpected API response data type. Expected {}, received {}'

    def __init__(self, recieved: Any, expected: type) -> None:
        super().__init__(
            self._PREFIX.format(str(expected), str(type(recieved)))
        )



"""
Amatino API Python Bindings
Entity Module
Author: hugh@amatino.io
"""
from amatino import Session
from amatino.region import Region
from amatino.user import User
from amatino.permissions_graph import PermissionsGraph
from amatino.internal.api_request import ApiRequest
from amatino.internal.data_package import DataPackage
from amatino.internal.http_method import HTTPMethod
from amatino.internal.url_parameters import UrlParameters
from amatino.internal.constrained_string import ConstrainedString
from amatino.internal.encodable import Encodable
from amatino.amatino_error import AmatinoError
from typing import TypeVar
from typing import Optional
from typing import Type
from typing import Dict
from typing import Any
from amatino.internal.immutable import Immutable

T = TypeVar('T', bound='Entity')


class Entity:
    """
    An Amatino entity is a economic unit to be described by accounting
    information. Entities are described by accounts, transactions, and entries.

    An example of an entity might be a legal company, a consolidated group
    of companies, a project, or even a person.
    """
    PATH = '/entities'
    MAX_NAME_LENGTH = 1024
    MAX_DESCRIPTION_LENGTH = 4096

    def __init__(
        self,
        session: Session,
        entity_id: str,
        name: str,
        description: str,
        region_id: int,
        owner_id: int,
        active: bool,
        permissions_graph: PermissionsGraph
    ) -> None:

        self._session = session
        self._entity_id = entity_id
        self._name = name
        self._description = description
        self._region_id = region_id
        self._owner_id = owner_id
        self._active = active
        self._permissions_graph = permissions_graph

        return

    session = Immutable(lambda s: s._session)
    id_ = Immutable(lambda s: s._entity_id)
    name = Immutable(lambda s: s._name)
    description = Immutable(lambda s: s._description)
    region_id = Immutable(lambda s: s._region_id)
    owner_id = Immutable(lambda s: s._owner_id)
    active = Immutable(lambda s: s._active)
    permissions_graph = Immutable(lambda s: s._permissions_graph)

    @classmethod
    def create(
        cls: Type[T],
        session: Session,
        name: str,
        description: Optional[str],
        region: Optional[Region] = None
    ) -> T:

        new_entity_arguments = cls.CreateArguments(
            name,
            description,
            region
        )

        request_data = DataPackage.from_object(new_entity_arguments)

        request = ApiRequest(
            Entity.PATH,
            HTTPMethod.POST,
            session,
            request_data,
            None,
            False
        )

        created_entity = cls._decode(request.response_data, session)

        return created_entity

    @classmethod
    def _decode(cls: Type[T], data: list, session: Session) -> T:
        """
        Return an Entity instance decoded from API response data
        """
        assert isinstance(session, Session)

        if not isinstance(data, list):
            raise AmatinoError('Unexpected response format: ' + str(type(data)))

        if len(data) < 1:
            raise AmatinoError('Response list unexpectedly empty')

        raw_entity = data[0]

        if not isinstance(raw_entity, dict):
            raise AmatinoError('Unexpected response format')
        try:
            entity = cls(
                session=session,
                entity_id=raw_entity['entity_id'],
                name=raw_entity['name'],
                description=raw_entity['description'],
                region_id=raw_entity['storage_region'],
                owner_id=raw_entity['owner'],
                active=raw_entity['active'],
                permissions_graph=PermissionsGraph(
                    raw_entity['permissions_graph']
                )
            )
        except KeyError:
            raise AmatinoError('Unexpected response format, missing a key')

        return entity

    @classmethod
    def retrieve(
        cls: Type[T],
        session: Session,
        entity_id: str
    ) -> T:
        if not isinstance(session, Session):
            raise TypeError('session must be of type `Session`')

        if not isinstance(entity_id, str):
            raise TypeError('entity_id must be of type `str`')

        url_parameters = UrlParameters(entity_id=entity_id)

        request = ApiRequest(
            path=Entity.PATH,
            method=HTTPMethod.GET,
            credentials=session,
            data=None,
            url_parameters=url_parameters,
            debug=False
        )

        entity = cls._decode(request.response_data, session)

        return entity

    def update(
        self,
        name: Optional[str] = None,
        description: Optional[str] = None,
        owner: Optional[User] = None,
        permissions_graph: Optional[PermissionsGraph] = None
    ) -> 'Entity':
        """
        Modify data describing this Entity. Returns this Entity, the Entity
        instance is not modified in place.
        """

        owner_id = None
        if owner is not None:
            owner_id = owner.id_

        update_arguments = Entity.UpdateArguments(
            self,
            name=name,
            description=description,
            owner_id=owner_id,
            permissions_graph=permissions_graph
        )

        data_package = DataPackage.from_object(data=update_arguments)

        request = ApiRequest(
            path=Entity.PATH,
            method=HTTPMethod.PUT,
            credentials=self._session,
            data=data_package,
            url_parameters=None
        )

        updated_entity = Entity._decode(request.response_data, self.session)

        return updated_entity

    def delete(self) -> None:
        """
        Destroy this Entity. Deleted entities can be recovered if necessary.
        Returns None, the Entity is updated-in-place.
        """
        raise NotImplementedError

    def restore(self) -> None:
        """
        Restore this Entity from a deleted state. Returns None, the Entity
        is updated-in-place.
        """
        raise NotImplementedError

    class _Description(ConstrainedString):
        def __init__(self, description: str) -> None:
            super().__init__(
                description,
                'description',
                Entity.MAX_DESCRIPTION_LENGTH
            )
            return

    class _Name(ConstrainedString):
        def __init__(self, name: str) -> None:
            super().__init__(
                name,
                'name',
                Entity.MAX_NAME_LENGTH
            )
            return

    class UpdateArguments(Encodable):
        def __init__(
            self,
            entity: T,
            name: Optional[str] = None,
            description: Optional[str] = None,
            owner_id: Optional[int] = None,
            permissions_graph: Optional[PermissionsGraph] = None
        ) -> None:

            if not isinstance(entity, Entity):
                raise TypeError('entity must be of type `Entity`')

            self._entity = entity

            if name:
                self._name = Entity._Name(name).serialise()
            else:
                self._name = entity.name

            if description:
                self._description = Entity._Description.serialise()
            else:
                self._description = entity._description

            if owner_id:
                if not isinstance(owner_id, int):
                    raise TypeError('owner_id must be of type `int`')
                self._owner_id = owner_id
            else:
                self._owner_id = entity.owner_id

            if permissions_graph:
                if not isinstance(permissions_graph, PermissionsGraph):
                    raise TypeError(
                        'graph must be of type `PermissionsGraph`'
                    )
                self._permissions_graph = permissions_graph
            else:
                self._permissions_graph = entity.permissions_graph

            return

        def serialise(self) -> Dict[str, Any]:
            data = {
                'name': self._name,
                'description': self._description,
                'entity_id': self._entity.id_,
                'owner': self._owner_id,
                'permissions_graph': self._permissions_graph.serialise()
            }
            return data

    class CreateArguments(Encodable):
        def __init__(
            self,
            name: str,
            description: Optional[str],
            region: Optional[Region] = None
        ) -> None:

            self._name = ConstrainedString(
                name,
                'name',
                Entity.MAX_NAME_LENGTH
            )

            self._description = ConstrainedString(
                description or '',
                'description',
                Entity.MAX_DESCRIPTION_LENGTH
            )

            if region is not None and not isinstance(region, Region):
                raise TypeError('region must be of type `Region`')

            self._region = region

            return

        def serialise(self) -> Dict[str, Any]:
            region_id = None
            if isinstance(self._region, Region):
                region_id = self._region.id_

            data = {
                'name': self._name.serialise(),
                'description': self._description.serialise(),
                'region_id': region_id
            }

            return data



from amatino.tests.derived.ledger import LedgerTest
from amatino.tests.derived.recursive_ledger import RecursiveLedgerTest
from amatino.tests.derived.balance import BalanceTest
from amatino.tests.derived.recursive_balance import RecursiveBalanceTest
from amatino.tests.derived.performance import PerformanceTest
from amatino.tests.derived.position import PositionTest



"""
Amatino API Python Bindings
Credentials Module
Author: hugh@amatino.io
"""


class Credentials:
    """
    Abstract class defining an interface for classes that may be used as
    credentials when authenticating and authorising requests to the Amatino
    API.
    """
    api_key = NotImplemented
    session_id = NotImplemented



"""
Amatino API Python Bindings
Test Module
Author: hugh@amatino.io

Base class for tests or, when executed as __main__, the entrypoint for the test
sequence.
"""
from urllib.error import HTTPError
from typing import Optional
from typing import Any
from os import environ
import traceback


ENVIRONMENT_HELP_MESSAGE = """
Amatino's test suite requires a few environment variables to run. It looks like
you are missing one or more of them. The required variables are:

AMATINO_TEST_USER_ID - Integer user id
AMATINO_TEST_EMAIL - String user email
AMATINO_TEST_SECRET - String user passphrase (take care not to save on disk!)

These three variable must be associated with a valid Amatino user account, with
active billing.

You can add the variables to your environment using the `export` command. For
example:

$ export AMATINO_TEST_EMAIL=some.cool@email.org

Once you've added the required variables, try running the test suite again.
"""


class Test:
    """
    Abstract class offering base-level functionality to tests.
    """
    def __init__(self, test_name: str) -> None:

        if not isinstance(test_name, str):
            raise TypeError('test_name must be of type `str`')

        try:
            raw_user_id = environ['AMATINO_TEST_USER_ID']
            email = environ['AMATINO_TEST_EMAIL']
            secret = environ['AMATINO_TEST_SECRET']
        except KeyError as error:
            print(ENVIRONMENT_HELP_MESSAGE)
            quit()

        if raw_user_id is None:
            raise RuntimeError('AMATINO_TEST_USER_ID env variable required')

        try:
            user_id = int(raw_user_id)
        except Exception as error:
            raise RuntimeError('AMATINO_TEST_USER_ID string must hold integer')

        if email is None:
            raise RuntimeError('AMATINO_TEST_EMAIL env variable required')

        if secret is None:
            raise RuntimeError('AMATINO_TEST_SECRET env variable required')

        self.user_id = user_id
        self.email = email
        self.secret = secret

        self._name = test_name
        self._note = None
        self._passed = None

        return

    def execute(self) -> None:
        """
        Run this test, recording a fail() or pass() for all possible execution
        paths such that pass or fail result is recorded at test conclusion.
        """
        raise NotImplementedError

    def record_success(self, note: str = None) -> None:
        """
        Assert that the test has finished and its conditions were met
        """
        self._record_result(True, note)
        return

    def record_failure(self, note: Optional[Any] = None) -> None:
        """
        Assert that the test has finished and its conditions were not met
        """
        self._record_result(False, note)
        return

    def _record_result(self, result: bool, note: Optional[str]) -> None:
        """
        Record an assertion of pass or failure
        """
        assert isinstance(result, bool)

        if self._passed is not None:
            raise RuntimeError('Attempt to pass/fail a completed test')
        self._passed = result

        if isinstance(note, HTTPError):
            self._note = 'An HTTP error occured: ' + str(note.code)
            try:
                self._note += '. The API returned the following:\n'
                self._note += '       ' + note.read().decode('utf-8')
                self._note += '\n Traceback follows:'
                self._note += '\n ' + traceback.format_exc()
            except Exception:
                pass
            return

        if isinstance(note, Exception):
            self._note = traceback.format_exc()
            return

        self._note = note

        return

    def report(self) -> str:
        """
        Return a string describing the outcome of this test.
        """
        if self._passed is None:
            raise RuntimeError('Cannot report on incomplete test')

        report = '[FAIL] '
        if self._passed:
            report = '[PASS] '
        report += self._name

        if self._note is not None:
            report += '\n       ' + str(self._note)

        return report



from amatino.session import Session
from amatino.entity import Entity
from amatino.account import Account
from amatino.am_type import AMType
from amatino.global_unit import GlobalUnit
from amatino.custom_unit import CustomUnit
from amatino.transaction import Transaction
from amatino.side import Side
from amatino.entry import Entry
from amatino.am_type import AMType
from amatino.ledger import Ledger
from amatino.recursive_ledger import RecursiveLedger
from amatino.ledger_row import LedgerRow
from amatino.user import User
from amatino.balance import Balance
from amatino.recursive_balance import RecursiveBalance
from amatino.performance import Performance
from amatino.position import Position
from amatino.tree_node import TreeNode



"""
Amatino API Python Bindings
Recursive Balance Module
Author: hugh@blinkybeach.com
"""
from amatino.internal.immutable import Immutable
from amatino.internal.balance_protocol import BalanceProtocol


class RecursiveBalance(BalanceProtocol):
    """
    A Recursive Balance represents the sum total value of all Entries party to
    an Account, and all of that Account's children.
    """

    PATH = '/accounts/balance/recursive'



"""
Amatino API Python Bindings
Account Module
Author: hugh@amatino.io
"""
from amatino.entity import Entity
from amatino.session import Session
from amatino.am_type import AMType
from amatino.global_unit import GlobalUnit
from amatino.custom_unit import CustomUnit
from amatino.denomination import Denomination
from amatino.color import Color
from typing import TypeVar
from typing import Optional
from typing import Type
from typing import Any
from typing import List
from amatino.internal.immutable import Immutable
from amatino.internal.encodable import Encodable
from amatino.internal.api_request import ApiRequest
from amatino.internal.constrained_string import ConstrainedString
from amatino.internal.data_package import DataPackage
from amatino.internal.http_method import HTTPMethod
from amatino.internal.url_parameters import UrlParameters
from amatino.internal.url_target import UrlTarget
from amatino.api_error import ApiError
from amatino.missing_key import MissingKey
from amatino.denominated import Denominated

T = TypeVar('T', bound='Account')


class Account(Denominated):
    """
    An Amatino Account is collection of related economic activity. For example,
    an Account might represent a bank account, income from a particular client,
    or company equity. Many Accounts together compose an Entity.
    """
    _PATH = '/accounts'
    MAX_DESCRIPTION_LENGTH = 1024
    MAX_NAME_LENGTH = 1024
    _URL_KEY = 'account_id'

    def __init__(
        self,
        session: Session,
        entity: Entity,
        account_id: int,
        name: str,
        am_type: AMType,
        description: str,
        parent_account_id: int,
        global_unit_id: Optional[int],
        custom_unit_id: Optional[int],
        counterparty_id: Optional[str],
        color: Color
    ) -> None:

        self._session = session
        self._entity = entity
        self._id = account_id
        self._name = name
        self._am_type = am_type
        self._description = description
        self._parent_account_id = parent_account_id
        self._global_unit_id = global_unit_id
        self._custom_unit_id = custom_unit_id
        self._counterparty_id = counterparty_id
        self._color = color

        self._cached_denomination = None

        return

    session = Immutable(lambda s: s._session)
    entity = Immutable(lambda s: s._entity)
    id_ = Immutable(lambda s: s._id)
    name = Immutable(lambda s: s._name)
    am_type = Immutable(lambda s: s._am_type)
    description = Immutable(lambda s: s._description)
    global_unit_id = Immutable(lambda s: s._global_unit_id)
    custom_unit_id = Immutable(lambda s: s._custom_unit_id)
    counterparty_id = Immutable(lambda s: s._counterparty_id)
    color = Immutable(lambda s: s._color)
    parent_id = Immutable(lambda s: s._parent_account_id)
    parent = Immutable(lambda s: s._parent())

    @classmethod
    def create(
        cls: Type[T],
        session: Session,
        entity: Entity,
        name: str,
        am_type: AMType,
        denomination: Denomination,
        description: Optional[str] = None,
        parent: Optional[T] = None,
        counter_party: Optional[Entity] = None,
        color: Optional[Color] = None
    ) -> T:

        arguments = Account.CreateArguments(
            name,
            description,
            am_type,
            parent,
            denomination,
            counter_party,
            color
        )

        data = DataPackage.from_object(arguments)
        parameters = UrlParameters(entity_id=entity.id_)

        request = ApiRequest(
            path=Account._PATH,
            method=HTTPMethod.POST,
            credentials=session,
            data=data,
            url_parameters=parameters
        )

        account = cls._decode(session, entity, request.response_data)

        return account

    @classmethod
    def retrieve(
        cls: Type[T],
        session: Session,
        entity: Entity,
        account_id: int
    ) -> T:
        """
        Return an existing Account
        """
        target = UrlTarget.from_integer(key=Account._URL_KEY, value=account_id)
        url_parameters = UrlParameters(entity_id=entity.id_, targets=[target])

        request = ApiRequest(
            path=Account._PATH,
            method=HTTPMethod.GET,
            credentials=session,
            data=None,
            url_parameters=url_parameters
        )

        account = cls._decode(session, entity, request.response_data)

        return account

    def update(
        self: T,
        name: Optional[str] = None,
        am_type: Optional[AMType] = None,
        parent: Optional[T] = None,
        description: Optional[str] = None,
        denomination: Optional[Denomination] = None,
        counterparty: Optional[Entity] = None,
        color: Optional[Color] = None
    ) -> 'Account':
        """
        Update this Account with new metadata.
        """

        arguments = Account.UpdateArguments(
            self,
            name,
            am_type,
            parent,
            description,
            denomination,
            counterparty,
            color
        )

        data = DataPackage.from_object(arguments)
        parameters = UrlParameters(entity_id=self.entity.id_)

        request = ApiRequest(
            path=Account._PATH,
            method=HTTPMethod.PUT,
            credentials=self.session,
            data=data,
            url_parameters=parameters
        )

        account = Account._decode(
            self.session,
            self.entity,
            request.response_data
        )

        if account.id_ != self.id_:
            raise ApiError('Returned Account ID does not match request ID')

        return account

    def delete(self):
        raise NotImplementedError

    def _denomination(self) -> Denomination:
        """Return the Denomination of this account"""
        if self._cached_denomination is not None:
            return self._cached_denomination
        if self._global_unit_id is not None:
            denomination = GlobalUnit.retrieve(
                self.session,
                self._global_unit_id
            )
        else:
            denomination = CustomUnit.retrieve(
                self.entity,
                self.session,
                self._custom_unit_id
            )
        assert isinstance(denomination, Denomination)
        self._cached_denomination = denomination
        return denomination

    class CreateArguments(Encodable):
        def __init__(
            self,
            name: str,
            description: Optional[str],
            am_type: AMType,
            parent: Optional[T],
            denomination: Denomination,
            counter_party: Optional[Entity],
            color: Optional[Color]
        ) -> None:

            self._name = Account._Name(name)
            if description is None:
                description = ''
            self._description = Account._Description(description)

            if not isinstance(am_type, AMType):
                raise TypeError('am_type must be of type `AMType`')

            self._type = am_type
            self._parent = Account._Parent(parent)
            self._global_unit = Account._GlobalUnit(denomination)
            self._custom_unit = Account._CustomUnit(denomination)

            if counter_party and not isinstance(counter_party, Entity):
                raise TypeError('counter_party must be of type `Entity`')

            self._counterparty = counter_party

            if color and not isinstance(color, Color):
                raise TypeError('color must be of type `Color`')

            self._color = color

            return

        def serialise(self) -> Any:

            counterparty_id = None
            if self._counterparty:
                counterparty_id = self._counterparty.id_

            color_code = None
            if self._color:
                color_code = str(self._color)

            data = {
                'name': self._name.serialise(),
                'type': self._type.value,
                'description': self._description.serialise(),
                'parent_account_id': self._parent.serialise(),
                'counterparty_entity_id': counterparty_id,
                'global_unit_id': self._global_unit.serialise(),
                'custom_unit_id': self._custom_unit.serialise(),
                'colour': color_code
            }
            return data

    @classmethod
    def _decode(
        cls: Type[T],
        session: Session,
        entity: Entity,
        data: List[dict]
    ) -> T:

        return cls._decode_many(session, entity, data)[0]

    @classmethod
    def _decode_many(
        cls: Type[T],
        session: Session,
        entity: Entity,
        data: List[dict]
    ) -> List[T]:

        if not isinstance(data, list):
            raise ApiError('Unexpected non-list data returned')

        if len(data) < 1:
            raise ApiError('Unexpected empty response data')

        def decode(data: dict) -> T:
            if not isinstance(data, dict):
                raise ApiError('Unexpected non-dict data returned')
            try:
                account = cls(
                    session=session,
                    entity=entity,
                    account_id=data['account_id'],
                    name=data['name'],
                    am_type=AMType(data['type']),
                    description=data['description'],
                    parent_account_id=data['parent_account_id'],
                    global_unit_id=data['global_unit_id'],
                    custom_unit_id=data['custom_unit_id'],
                    counterparty_id=data['counterparty_entity_id'],
                    color=Color.from_hex_string(data['colour'])
                )
            except KeyError as error:
                raise MissingKey(error.args[0])

            return account

        accounts = [decode(a) for a in data]

        return accounts

    def _parent(self) -> Optional['Account']:
        """Return this Account's parent, if it has one"""
        if self.parent_id is None:
            return None
        assert isinstance(self.parent_id, int)
        return Account.retrieve(self.session, self.entity, self.parent_id)

    class UpdateArguments(Encodable):
        def __init__(
            self,
            account: T,
            name: Optional[str],
            am_type: Optional[AMType],
            parent: Optional[T],
            description: Optional[str],
            denomination: Optional[Denomination],
            counterparty: Optional[Entity],
            color: Optional[Color]
        ) -> None:

            if not isinstance(account, Account):
                raise TypeError('account must be of type `Account`')

            self._account_id = account.id_

            if not name:
                name = account.name
            assert isinstance(name, str)
            self._name = Account._Name(name)

            if am_type:
                if not isinstance(am_type, AMType):
                    raise TypeError('am_type must be of type `AMType`')
            else:
                am_type = account.am_type

            self._type = am_type

            if not description:
                description = account.description
            self._description = Account._Description(description)

            if not parent:
                self._parent_id = account.parent_id
            else:
                self._parent_id = Account._Parent(parent).serialise()

            if denomination:
                self._global_unit_id = Account._GlobalUnit(
                    denomination
                ).serialise()
                self._custom_unit_id = Account._CustomUnit(
                    denomination
                ).serialise()
            else:
                self._global_unit_id = account.global_unit_id
                self._custom_unit_id = account.custom_unit_id

            if counterparty:
                if not isinstance(counterparty, Entity):
                    raise TypeError('counterparty must be of type `Entity`')
                self._counterparty_id = counterparty.id_
            else:
                self._counterparty_id = account.counterparty_id

            if color:
                if not isinstance(color, Color):
                    raise TypeError('color must be of type `Color`')
                self._color = color
            else:
                self._color = account.color

            return

        def serialise(self) -> Any:

            data = {
                'name': self._name.serialise(),
                'account_id': self._account_id,
                'description': self._description.serialise(),
                'type': self._type.value,
                'parent_account_id': self._parent_id,
                'global_unit_id': self._global_unit_id,
                'custom_unit_id': self._custom_unit_id,
                'colour': self._color.serialise(),
                'counterparty_entity_id': self._counterparty_id
            }
            return data

    class _Name(Encodable):
        def __init__(self, string: str) -> None:
            if not isinstance(string, str):
                raise TypeError('name must be of type `str`')
            self._name = ConstrainedString(
                string,
                'name',
                Account.MAX_NAME_LENGTH
            )
            return

        def serialise(self) -> str:
            return str(self._name)

    class _Description(Encodable):
        def __init__(self, string: Optional[str]) -> None:
            if string is not None and not isinstance(string, str):
                raise TypeError('description must be of type `str` or None')
            self._description = ConstrainedString(
                string,
                'description',
                Account.MAX_DESCRIPTION_LENGTH
            )
            return

        def serialise(self) -> str:
            return str(self._description)

    class _Parent(Encodable):
        def __init__(self, parent: Optional[T]) -> None:
            if parent is not None and not isinstance(parent, Account):
                raise TypeError('parent must be of type `Account`')
            self._parent_id = None
            if parent is not None:
                self._parent_id = parent.id_
            return

        def serialise(self) -> Optional[int]:
            return self._parent_id

    class _CustomUnit(Encodable):
        def __init__(self, denomination: Denomination) -> None:

            if not isinstance(denomination, Denomination):
                raise TypeError('denomination must be of type `Denomination`')

            self._custom_unit_id = None

            if isinstance(denomination, CustomUnit):
                self._custom_unit_id = denomination.id_

            return

        def serialise(self) -> Optional[int]:
            return self._custom_unit_id

    class _GlobalUnit(Encodable):
        def __init__(self, denomination: Denomination) -> None:

            if not isinstance(denomination, Denomination):
                raise TypeError('denomination must be of type `Denomination`')

            self._global_unit_id = None

            if isinstance(denomination, GlobalUnit):
                self._global_unit_id = denomination.id_

            return

        def serialise(self) -> Optional[int]:
            return self._global_unit_id



"""
Amatino API Python Bindings
Position Test Module
author: hugh@blinkybeach.com
"""
from amatino.tests.primary.account import AccountTest
from amatino import Account
from amatino import Transaction
from amatino import Position
from amatino import Entity
from amatino import GlobalUnit
from amatino import TreeNode
from amatino import AMType
from amatino import Entry
from amatino import Side
from decimal import Decimal
from datetime import datetime
from datetime import timedelta

NAME = 'Retrieve a Position'


class PositionTest(AccountTest):
    """Test the Position object"""

    def __init__(self, name=NAME) -> None:
        super().__init__(name)
        return

    def execute(self) -> None:

        try:

            asset = self.create_account(
                AMType.asset,
                'Test Asset'
            )

            liability = self.create_account(
                AMType.liability,
                'Test Liability'
            )

            liability_child = Account.create(
                session=self.session,
                entity=self.entity,
                name='Test liability child',
                am_type=AMType.liability,
                denomination=liability.denomination,
                parent=liability
            )

            Transaction.create(
                self.session,
                self.entity,
                datetime.utcnow(),
                [
                    Entry(Side.debit, Decimal('400'), asset),
                    Entry(Side.credit, Decimal('400'), liability)
                ],
                self.usd,
                'Test transaction 1'
            )

            Transaction.create(
                self.session,
                self.entity,
                datetime.utcnow(),
                [
                    Entry(Side.debit, Decimal('700'), asset),
                    Entry(Side.credit, Decimal('200'), liability),
                    Entry(Side.credit, Decimal('500'), liability_child)
                ],
                self.usd,
                'Test transaction 2'
            )

            Transaction.create(
                self.session,
                self.entity,
                datetime.utcnow() + timedelta(days=4),
                [
                    Entry(Side.debit, Decimal('1'), asset),
                    Entry(Side.credit, Decimal('1'), liability)
                ],
                self.usd,
                'Test transaction 3'
            )

            position = Position.retrieve(
                entity=self.entity,
                balance_time=datetime.utcnow(),
                denomination=self.usd
            )

            assert isinstance(position, Position)
            assert position.has_assets
            assert position.has_liabilities
            assert not position.has_equities
            assert isinstance(position.denomination, GlobalUnit)
            assert isinstance(position.balance_time, datetime)
            assert isinstance(position.generated_time, datetime)
            assert isinstance(position.entity, Entity)

            asset_root = position.assets[0]
            liability_root = position.liabilities[0]

            assert isinstance(asset_root, TreeNode)
            assert isinstance(liability_root, TreeNode)
            assert liability_root.has_children
            assert not asset_root.has_children

            liability_depth_1 = liability_root.children[0]

            assert isinstance(liability_depth_1, TreeNode)
            assert not liability_depth_1.has_children

            assert asset_root.recursive_balance == Decimal('1100')
            assert asset_root.account_balance == Decimal('1100')
            assert liability_root.account_balance == Decimal('600')
            assert liability_root.recursive_balance == Decimal('1100')
            assert liability_depth_1.account_balance == Decimal('500')
            assert liability_depth_1.recursive_balance == Decimal('500')

        except Exception as error:
            self.record_failure(error)
            return

        self.record_success()



"""
Amatino API Python Bindings
API Request Module
Author: hugh@amatino.io

This module is intended to be private, used indirectly
by public classes, and should not be used directly.
"""
from amatino.internal.url_target import UrlTarget
from typing import Optional
from typing import List
from typing import Type
from typing import TypeVar

T = TypeVar('T', bound='UrlParameters')


class UrlParameters:
    """
    Private - Not intended to be used directly.

    An instance of url parameters to be included in a request
    to the Amatino Api
    """
    def __init__(
        self,
        entity_id: Optional[str] = None,
        targets: Optional[List[UrlTarget]] = None,
        raw_query_string: Optional[str] = None
    ) -> None:

        if raw_query_string is not None:
            if not isinstance(raw_query_string, str):
                raise TypeError('raw_query_string must be of type str')
            self._parameter_string = raw_query_string
            return

        if entity_id is not None and not isinstance(entity_id, str):
            raise TypeError("Entity must be of type str")

        if (
                targets is not None
                and (
                    not isinstance(targets, list)
                    or False in [isinstance(t, UrlTarget) for t in targets]
                )
        ):
            raise TypeError('Targets must be of type List[UrlTarget]')

        self._parameter_string = ''

        if entity_id is None and targets is None:
            return

        if entity_id is not None:
            self._parameter_string = '?entity_id=' + entity_id

        if targets is None:
            return

        remaining_targets = targets[:]

        if entity_id is None:
            self._parameter_string = '?' + str(targets[0])
            remaining_targets = targets[1:]

        for target in remaining_targets:
            self._parameter_string += '&' + str(target)

        return

    @classmethod
    def from_targets(cls: Type[T], targets: List[UrlTarget]) -> T:
        """Initialise UrlParameters with a set of UrlTargets only"""
        return cls(targets=targets)

    @classmethod
    def from_single_target(cls: Type[T], target: UrlTarget) -> T:
        """Initialise UrlParameters with a single UrlTarget"""
        return cls(targets=[target])

    def __str__(self):
        return self._parameter_string

    def parameter_string(self) -> str:
        """
        Return a string of url parameters suitable for inclusion
        in a request to the Amatino API.
        """
        return self._parameter_string



"""
Amatino API Python Bindings
API Error Module
Author: hugh@amatino.io
"""
from amatino.amatino_error import AmatinoError


class ApiError(AmatinoError):
    """
    An error caused by the Amatino API returning data that Amatino Python could
    not parse, or interpreted as being invalid. This error type indicates
    that there is a bug in either the Amatino API or Amatino Python.
    """
    pass



"""
Amatino API Python Bindings
API Request Module
Author: hugh@amatino.io

This module is intended to be private, used indirectly
by public classes, and should not be used directly.
"""
from json import loads
from urllib.request import Request
from urllib.request import urlopen
from urllib.error import HTTPError
from amatino.internal.credentials import Credentials
from amatino.internal.data_package import DataPackage
from amatino.internal.url_parameters import UrlParameters
from amatino.internal.request_headers import RequestHeaders
from amatino.internal.http_method import HTTPMethod
from typing import Optional


class ApiRequest:
    """
    Private - Not intended to be used directly.

    An instance of an http request to the Amatino API.
    """

    #_ENDPOINT = 'https://api.amatino.io'
    #_ENDPOINT = 'http://172.16.101.154'
    _ENDPOINT = 'http://127.0.0.1:5000'
    _DEBUG_ENDPOINT = 'http://127.0.0.1:5000'
    _TIMEOUT = 10

    def __init__(
        self,
        path: str,
        method: HTTPMethod,
        credentials: Optional[Credentials] = None,
        data: Optional[DataPackage] = None,
        url_parameters: Optional[UrlParameters] = None,
        debug: bool = False
    ) -> None:

        self.response_data = None

        if credentials is not None:
            assert isinstance(credentials, Credentials)

        if data is not None:
            assert isinstance(data, DataPackage)
            request_data = data.as_json_bytes()
        else:
            request_data = None

        if url_parameters is not None:
            assert isinstance(url_parameters, UrlParameters)

        if debug is False:
            url = self._ENDPOINT
        else:
            url = self._DEBUG_ENDPOINT

        url += path

        if url_parameters is not None:
            url += url_parameters.parameter_string()

        headers = RequestHeaders(path, credentials, data)
        request = Request(
            url=url,
            data=request_data,
            headers=headers.dictionary(),
            method=method.value,
        )
        try:
            self._response = urlopen(request, timeout=self._TIMEOUT)
        except HTTPError as error:
            # Insert error handling
            raise error

        self.response_data = loads(self._response.read().decode('utf-8'))

        return



"""
Amatino API Python Bindings
Signature Module
Author: hugh@amatino.io

This module is intended to be private, used indirectly by public classes, and
should not be used directly.
"""
from hashlib import sha512
from base64 import b64encode
import hmac
import time
import json


class Signature:
    """
    Private - Not intended to be used directly.

    An HMAC signature computation
    """
    def __init__(
        self,
        api_key: str,
        path: str,
        json_data = None
        ):

        assert isinstance(api_key, str)
        assert isinstance(path, str)
        if json_data is not None:
            assert isinstance(json_data, dict) or isinstance(json_data, list)
        
        if json_data is not None:
            json_string = json.dumps(json_data, separators=(',', ':'))
            assert isinstance(json_string, str)
            if json_string == 'null':
                json_string = None
        else:
            json_string = None

        encoded_key = api_key.encode('utf-8')

        timestamp = str(int(time.time()))

        if json_string is not None:
            message = timestamp + path + json_string
        else:
            message = timestamp + path

        encoded_message = message.encode('utf-8')

        digest = hmac.new(encoded_key, encoded_message, sha512).digest()
        self._signature = b64encode(digest).decode()

        return

    def string(self) -> str:
        """
        Return a string signature
        """
        assert isinstance(self._signature, str)
        return self._signature



"""
Amatino API Python Bindings
Entry Module
Author: hugh@amatino.io
"""
from amatino.side import Side
from amatino.account import Account
from amatino.internal.constrained_string import ConstrainedString
from amatino.internal.encodable import Encodable
from amatino.internal.immutable import Immutable
from decimal import Decimal
from typing import Dict
from typing import Any
from typing import TypeVar
from typing import Optional
from typing import Type
from typing import List

T = TypeVar('T', bound='Entry')


class Entry(Encodable):
    """
    Entries compose Transactions. An individual entry allocates some value to
    an Account as either one of the fundamental Sides: a debit or a credit.
    All together, those debits and credits will add up to zero, satisfying the
    fundamental double-entry accounting equality.
    """
    MAX_DESCRIPTION_LENGTH = 1024

    def __init__(
        self,
        side: Side,
        amount: Decimal,
        account: Optional[Account] = None,
        description: Optional[str] = None,
        account_id: Optional[int] = None
    ) -> None:

        if not isinstance(side, Side):
            raise TypeError('side must be of type `Side`')

        if not isinstance(amount, Decimal):
            raise TypeError('amount must be of type `Decimal`')

        self._side = side
        if account_id is not None:
            assert isinstance(account_id, int)
            self._account_id = account_id
        else:
            if not isinstance(account, Account):
                raise TypeError('account must be of type `Account`')
            self._account_id = account.id_
        self._amount = amount
        self._description = Entry._Description(description)

        return

    side = Immutable(lambda s: s._side)
    account_id = Immutable(lambda s: s._account_id)
    amount = Immutable(lambda s: s._amount)
    description = Immutable(lambda s: s._description)

    def serialise(self) -> Dict[str, Any]:
        data = {
            'account_id': self._account_id,
            'amount': str(self._amount),
            'description': self._description.serialise(),
            'side': self._side.value
        }
        return data

    class _Description(Encodable):
        def __init__(self, string: Optional[str]) -> None:
            if string is not None and not isinstance(string, str):
                raise TypeError('description must be of type `str` or None')
            if string is None:
                string = ''
            self._description = ConstrainedString(
                string,
                'description',
                Entry.MAX_DESCRIPTION_LENGTH
            )
            return

        def serialise(self) -> str:
            return str(self._description)

    @classmethod
    def create(
        cls: Type[T],
        side: Side,
        amount: Decimal,
        account: Account,
        description: Optional[str] = None
    ) -> T:

        return cls(side, amount, account=account, description=description)

    @classmethod
    def create_balanced_pair(
        cls: Type[T],
        debit_account: Account,
        credit_account: Account,
        amount: Decimal,
        description: Optional[str] = None
    ) -> List[T]:

        debit = cls(Side.debit, amount, debit_account, description)
        credit = cls(Side.credit, amount, credit_account, description)

        return [debit, credit]

    @classmethod
    def plug(
        cls: Type[T],
        account: Account,
        entries: List[T],
        description: Optional[str] = None
    ) -> Optional[T]:
        """
        Return an entry plugging the balance gap in a given set of Entries. Or,
        return None if the Entries already balance.
        """

        if False in [isinstance(e, Entry) for e in entries]:
            raise TypeError('Entries must be of type List[Entry]')

        debits = sum([e.amount for e in entries if e.side == Side.debit])
        credits_ = sum([e.amount for e in entries if e.side == Side.credit])

        if debits == credits_:
            return None

        if debits > credits_:
            plug_side = Side.credit
            amount = Decimal(debits - credits_)
        else:
            plug_side = Side.debit
            amount = Decimal(credits_ - debits)

        return cls(plug_side, amount, account, description)



"""
Amatino API Python Bindings
Session Module
Author: hugh@amatino.io
"""
from amatino.internal.session_create_arguments import NewSessionArguments
from amatino.internal.data_package import DataPackage
from amatino.internal.api_request import ApiRequest
from amatino.api_error import ApiError
from typing import TypeVar
from typing import Type
from typing import Any
from typing import Optional
from amatino.internal.immutable import Immutable
from amatino.internal.http_method import HTTPMethod
from amatino.internal.credentials import Credentials

T = TypeVar('T', bound='Session')


class Session(Credentials):
    """
    Sessions are the keys to the Amatino kingdom. All requests to the Amatino
    API, except those requests to create Sessions themselves, must include two
    HTTP headers: An integer session identifier, and a Hashed Message
    Authentication Code (HMAC) signed with a Session API Key. The Session object
    handles said header construction and HMAC signing for you behind the scenes.

    Creating a new Session is analogous to 'logging in', and deleting a Session
    with the delete() method is analogous to 'logging out'. Your application
    might wish to create multiple Sessions for a User. For example, one per
    device.
    """

    _PATH = '/session'

    def __init__(
        self,
        user_id: int = None,
        session_id: int = None,
        api_key: str = None
    ) -> None:

        self._api_key = api_key
        self._session_id = session_id
        self._user_id = user_id

        return

    api_key = Immutable(lambda s: s._api_key)
    session_id = Immutable(lambda s: s._session_id)
    user_id = Immutable(lambda s: s._user_id)
    id_ = Immutable(lambda s: s.session_id)

    @classmethod
    def create_with_email(cls: Type[T], email: str, secret: str) -> T:
        """
        Create a new session using an email / secret pair.
        """
        if not isinstance(email, str):
            raise TypeError('email must be of type `str`')

        if not isinstance(secret, str):
            raise TypeError('secret must be of type `str')

        return cls._create(None, email, secret)

    @classmethod
    def create_with_user_id(cls: Type[T], user_id: int, secret: str) -> T:
        """
        Create a new session using a user_id / secret pair
        """
        if not isinstance(user_id, int):
            raise TypeError('user_id must be of type `int`')

        if not isinstance(secret, str):
            raise TypeError('secret must be of type `str`')

        return cls._create(user_id, None, secret)

    @classmethod
    def _create(
        cls: Type[T],
        user_id: Optional[int],
        email: Optional[str],
        secret: str
    ) -> T:

        new_arguments = NewSessionArguments(secret, email, user_id)

        request_data = DataPackage(
            object_data=new_arguments,
            override_listing=True
        )

        request = ApiRequest(
            path=cls._PATH,
            method=HTTPMethod.POST,
            data=request_data
        )

        entity = cls._decode(request.response_data)

        return entity

    @classmethod
    def _decode(cls: Type[T], response_data: Any):

        if not isinstance(response_data, dict):
            raise ApiError('Unexpected non-dict type when decoding Session')

        try:
            api_key = response_data['api_key']
            session_id = response_data['session_id']
            user_id = response_data['user_id']
        except KeyError as error:
            message = 'Expected key "{key}" missing from response data'
            message.format(key=error.args[0])
            raise ApiError(message)

        entity = cls(user_id, session_id, api_key)

        return entity

    def delete(self):
        """
        Destroy this Session, such that its id and api_key are no
        longer valid for authenticating Amatino API requests. Analagous
        to 'logging out' the underlying User.
        """
        raise NotImplementedError



"""
Amatino API Python Bindings
Missing Key Module
Author: hugh@amatino.io
"""
from amatino.api_error import ApiError


class MissingKey(ApiError):
    _PREFIX = 'Expected key "{}" missing from response data'

    def __init__(self, key: str) -> None:
        super().__init__(self._PREFIX.format(key))



"""
Amatino API Python Bindings
Constraint Error Module
Author: hugh@amatino.io
"""
from amatino.amatino_error import AmatinoError


class ConstraintError(AmatinoError):
    """
    A class of errors thrown when input data violates some constraint. For
    example, an Entity name that is too long.
    """
    pass



"""
Amatino API Python Bindings
Entity Update Arguments
Author: hugh@amatino.io
"""
from amatino.internal.encodable import Encodable
from amatino.internal.entity_create_arguments import NewEntityArguments
from amatino.internal.constrained_string import ConstrainedString
from typing import Optional


class EntityUpdateArguments(Encodable):
    """
    A set of arguments suitable for provision to the Amatino API for the
    purpose of updating an existing Entity
    """
    def __init__(
        self,
        entity_id: str,
        name: str,
        description: str,
        owner_id: int,
        permissions_graph: Optional[dict]
    ) -> None:

        assert isinstance(entity_id, str)

        self._entity_id = entity_id

        self._name = ConstrainedString(
            name,
            'name',
            NewEntityArguments.MAX_NAME_LENGTH
        )

        self._description = ConstrainedString(
            description,
            'description',
            NewEntityArguments.MAX_DESCRIPTION_LENGTH
        )

        if not isinstance(owner_id, int):
            raise TypeError('owner_id must be of type `int`')

        self._owner_id = owner_id

        self._permissions_graph = permissions_graph

        if permissions_graph is None:
            return

        if not isinstance(permissions_graph, dict):
            raise TypeError('permissions_graph must be of type `dict`')

        if False in [isinstance(k, str) for k in permissions_graph]:
            raise TypeError('permissions_graph keys must be of type `str`')

        upg = permissions_graph
        if False in [isinstance(upg[k], dict) for k in upg]:
            raise TypeError('permissions_graph values must be of type `dict`')

        return

    def serialise(self):
        data = {
            'entity_id': self._entity_id,
            'name': str(self._name),
            'description': str(self._description),
            'owner_id': self._owner_id,
            'permissions_graph': self._permissions_graph
        }
        return data



"""
Amatino API Python Bindings
Session Test Module
Author: hugh@amatino.io
"""
from amatino.tests.test import Test
from amatino.session import Session
from typing import Optional
from typing import Any
from amatino.amatino_error import AmatinoError


class SessionTest(Test):
    """
    Test the Session ancillary object
    """
    TEST_ALL_SESSION_METHODS = True

    def __init__(self, name='Create a Session') -> None:

        super().__init__(name)

        self.session = None

        return

    def create_session(self) -> Optional[Any]:
        session = Session.create_with_email(
            self.email,
            self.secret
        )
        self.session = session
        return session

    def execute(self) -> None:

        try:
            session = Session.create_with_email(
                self.email,
                self.secret
            )
        except Exception as error:
            self.record_failure(error)
            return

        did_fail_to_set = False
        try:
            session.api_key = 0
        except AmatinoError:
            did_fail_to_set = True
            pass

        if did_fail_to_set is not True:
            self.record_failure('Property set did not raise exception')
            return

        try:
            session = Session.create_with_user_id(
                self.user_id,
                self.secret
            )
        except Exception as error:
            self.record_failure(error)
            return

        self.record_success()
        return



"""
Amatino API Python Bindings
Request Headers Module
Author: hugh@amatino.io

This module is intended to be private, used indirectly
by public classes, and should not be used directly.
"""
from amatino.internal.data_package import DataPackage
from amatino.internal.signature import Signature
from amatino.internal.credentials import Credentials


class RequestHeaders:
    """
    Private - Not intended to be used directly.

    An instance of HTTP headers for use in an API request.
    """
    _AGENT = 'Amatino Python'

    def __init__(
        self,
        path: str,
        credentials: Credentials = None,
        request_data: DataPackage = None
    ) -> None:

        self._headers = {'User-Agent': self._AGENT}

        if request_data is not None:
            self._headers['content-type'] = 'application/json'

        if credentials is None:
            return

        json_data = None
        if request_data is not None:
            json_data = request_data.as_object()

        signature = Signature(
            api_key=credentials.api_key,
            path=path,
            json_data=json_data
        )

        self._headers['X-Signature'] = signature.string()
        self._headers['X-Session-ID'] = credentials.session_id

        return

    def dictionary(self) -> dict:
        """
        Return request headers as a dict
        """
        assert isinstance(self._headers, dict)
        return self._headers



"""
Amatino API Python Bindings
Recursive Ledger Test Module
Author: hugh@amatino.io
"""
from datetime import datetime
from amatino.tests.primary.transaction import TransactionTest
from amatino import Account
from amatino import Transaction
from amatino import RecursiveLedger
from amatino import Side
from amatino import Entry
from decimal import Decimal

NAME = 'Retrieve a RecursiveLedger'


class RecursiveLedgerTest(TransactionTest):
    """Test the RecursiveLedger object"""

    def __init__(self, name=NAME) -> None:

        super().__init__(name)
        return

    def execute(self) -> None:

        try:
            asset_child = Account.create(
                self.session,
                self.entity,
                'Test Asset Child',
                self.asset.am_type,
                self.asset.denomination,
                parent=self.asset
            )

            Transaction.create(
                self.session,
                self.entity,
                datetime.utcnow(),
                [
                    Entry(Side.debit, Decimal(5), self.asset),
                    Entry(Side.credit, Decimal(5), self.liability)
                ],
                self.asset.denomination
            )

            Transaction.create(
                self.session,
                self.entity,
                datetime.utcnow(),
                [
                    Entry(Side.debit, Decimal(5), asset_child),
                    Entry(Side.credit, Decimal(5), self.liability)
                ],
                asset_child.denomination
            )

            ledger = RecursiveLedger.retrieve(
                self.session,
                self.entity,
                self.asset
            )
        except Exception as error:
            self.record_failure(error)
            return

        if not isinstance(ledger, RecursiveLedger):
            return_type = str(type(ledger))
            self.record_failure('Unexpected return type: ' + return_type)

        if len(ledger) != 2:
            self.record_failure('Ledger length unexpected: ' + str(len(ledger)))
            return

        if ledger.account_id != self.asset.id_:
            self.record_failure('Account IDs do not match')
            return

        self.record_success()

        return



"""
Amatino API Python Bindings
Encodable Module
Author: hugh@amatino.io

This module is intended to be private, used indirectly by public classes, and
should not be used directly.
"""
from typing import Any
import json


class Encodable:
    """
    Abstract class defining an interface for types that may be serialised
    to JSON.
    """
    def serialise(self) -> Any:
        """
        Return a version of this object in a serialisable form.
        """
        raise NotImplementedError

    def encode_to_json(self) -> str:
        """
        Return a JSON string version of this object
        """
        return json.dumps(self.serialise())



"""
Amatino API Python Bindings
Data Package Module
Author: hugh@amatino.io

This module is intended to be private, used indirectly by public classes, and
should not be used directly.
"""
import json
from amatino.internal.encodable import Encodable
from typing import TypeVar
from typing import Any
from typing import List
from typing import Type

T = TypeVar('T', bound='DataPackage')


class DataPackage:
    """
    Private - Not intended to be used directly.

    Abstract class providing JSON data packing capabilities to concrete classes.
    """
    def __init__(
        self,
        list_data: List[Encodable] = None,
        object_data: Encodable = None,
        override_listing: bool = False,
        raw_data: Any = None
    ) -> None:

        self._data = None

        if raw_data is not None:
            self._data = raw_data
            return

        assert isinstance(override_listing, bool)
        assert (list_data is None or object_data is None)

        if list_data is not None:
            assert isinstance(list_data, list)
            assert False not in [
                isinstance(e, Encodable) for e in list_data
            ]
            assert override_listing is False
            self._data = [e.serialise() for e in list_data]

        if object_data is not None:
            assert isinstance(object_data, Encodable)
            if override_listing is True:
                self._data = object_data.serialise()
            else:
                self._data = [object_data.serialise()]

        assert self._data is not None
        return

    @classmethod
    def from_object(
        cls: Type[T],
        data: Encodable,
        override_listing: bool = False
    ) -> T:
        """
        Return a new instance of DataPackage initialised with a single Encodable
        object
        """
        return cls(object_data=data, override_listing=override_listing)

    def as_json_bytes(self) -> bytes:
        """
        Return package arguments as JSON bytes suitable
        for inclusion in an HTTP request.
        """
        return json.dumps(self._data).encode('utf-8')

    def as_object(self):
        """
        Return the underlying object data, either a list
        or a dictionary.
        """
        return self._data



"""
Amatino API Python Bindings
Custom Unit Module
Author: hugh@amatino.io
"""
from amatino.session import Session
from amatino.entity import Entity
from amatino.global_unit import GlobalUnit
from amatino.denomination import Denomination
from typing import TypeVar
from typing import Optional
from typing import Type
from typing import List
from typing import Any
from typing import Dict
from amatino.internal.immutable import Immutable
from amatino.internal.constrained_string import ConstrainedString
from amatino.internal.constrained_integer import ConstrainedInteger
from amatino.internal.encodable import Encodable
from amatino.internal.url_target import UrlTarget
from amatino.internal.url_parameters import UrlParameters
from amatino.internal.api_request import ApiRequest
from amatino.internal.http_method import HTTPMethod
from amatino.internal.data_package import DataPackage
from amatino.api_error import ApiError

T = TypeVar('T', bound='CustomUnit')


class CustomUnit(Denomination):
    """
    Custom Units are units of account created by Amatino users. Their scope is
    limited to the Entity in which they are created. They can be used anywhere
    a Global unit would be used, allowing a user to denominate their
    Transactions and Accounts as they please.

    Custom Unit identifiers must be unique with reference to each other, but
    need not be so with reference to Global Units. Therefore, it is possible
    to create a Custom Unit implementation of a Global Unit - For example, a
    USD Custom Unit using a preferred source of foreign exchange rates.
    """
    MAX_DESCRIPTION_LENGTH = 1024
    MIN_CODE_LENGTH = 3
    MAX_CODE_LENGTH = 64
    MAX_NAME_LENGTH = 1024
    MAX_EXPONENT_VALUE = 6
    MIN_EXPONENT_VALUE = 0
    MAX_PRIORITY_VALUE = 10000
    MIN_PRIORITY_VALUE = -10000
    _URL_KEY = 'custom_unit_id'
    _PATH = '/custom_units'

    def __init__(
        self,
        session: Session,
        entity: Entity,
        id_: int,
        code: str,
        name: str,
        priority: int,
        description: str,
        exponent: int
    ) -> None:

        self._session = session
        self._entity = entity

        super().__init__(code, id_, name, priority, description, exponent)

        return

    session = Immutable(lambda s: s._session)
    entity = Immutable(lambda s: s._entity)

    @classmethod
    def create(
        cls: Type[T],
        entity: Entity,
        session: Session,
        name: str,
        code: str,
        exponent: int,
        description: Optional[str] = None,
        priority: Optional[int] = None
    ) -> T:

        arguments = CustomUnit.CreationArguments(
            code,
            name,
            exponent,
            priority,
            description
        )

        parameters = UrlParameters(entity_id=entity.id_)

        request = ApiRequest(
            path=CustomUnit._PATH,
            credentials=session,
            method=HTTPMethod.POST,
            data=DataPackage.from_object(arguments),
            url_parameters=parameters
        )

        return CustomUnit._decode(session, entity, request.response_data)

    @classmethod
    def retrieve(
        cls: Type[T],
        entity: Entity,
        session: Session,
        custom_unit_id: int
    ) -> T:

        target = UrlTarget.from_integer(cls._URL_KEY, custom_unit_id)
        parameters = UrlParameters(entity_id=entity.id_, targets=[target])

        request = ApiRequest(
            path=cls._PATH,
            method=HTTPMethod.GET,
            data=None,
            url_parameters=parameters,
            credentials=session
        )

        return cls._decode(session, entity, request.response_data)

    @classmethod
    def _retrieve(
        cls: Type[T],
        entity: Entity,
        session: Session,
        custom_unit_ids: List[int]
    ) -> T:

        if not isinstance(entity, Entity):
            raise TypeError('entity must be of type `Entity`')

        if not isinstance(session, Session):
            raise TypeError('session must be of type `Session')

        key = CustomUnit._URL_KEY
        targets = [UrlTarget(key, str(i)) for i in custom_unit_ids]

        url_parameters = UrlParameters(
            entity_id=entity.id_,
            targets=targets
        )

        request = ApiRequest(
            path=CustomUnit._PATH,
            method=HTTPMethod.GET,
            credentials=session,
            data=None,
            url_parameters=url_parameters
        )

        unit = cls._decode(
            session,
            entity,
            request.response_data
        )

        return unit

    @classmethod
    def _decode(
        cls: Type[T],
        session: Session,
        entity: Entity,
        data: Any
    ) -> T:
        return cls._decodeMany(session, entity, data)[0]

    @classmethod
    def _decodeMany(
        cls: Type[T],
        session: Session,
        entity: Entity,
        data: Any
    ) -> List[T]:

        assert isinstance(session, Session)
        assert isinstance(entity, Entity)

        if not isinstance(data, list):
            raise ApiError('Unexpected non-list data returned')

        if len(data) < 1:
            raise ApiError('Unexpected empty response data')

        def decode(data: dict) -> T:
            if not isinstance(data, dict):
                raise ApiError('Unexpected non-dict data returned')
            try:
                unit = cls(
                    session=session,
                    entity=entity,
                    id_=data['custom_unit_id'],
                    code=data['code'],
                    name=data['name'],
                    priority=data['priority'],
                    description=data['description'],
                    exponent=data['exponent']
                )
            except KeyError as error:
                message = 'Expected key "{key}" missing from response data'
                message.format(key=error.args[0])
                raise ApiError(message)

            return unit

        units = [decode(u) for u in data]

        return units

    def update(
        self,
        name: Optional[str] = None,
        code: Optional[str] = None,
        priority: Optional[int] = None,
        description: Optional[str] = None,
        exponent: Optional[str] = None
    ) -> None:
        """
        Replace existing Custom Unit data with supplied data. Parameters
        not supplied will default to existing values.
        """
        if name is None:
            name = self.name
        if code is None:
            code = self.code
        if priority is None:
            priority = self.priority
        if description is None:
            description = self.description
        if exponent is None:
            exponent = self.exponent

        assert isinstance(code, str)
        assert isinstance(priority, int)
        assert isinstance(description, str)
        assert isinstance(exponent, int)

        arguments = CustomUnit.UpdateArguments(
            self.id_,
            code,
            name,
            exponent,
            priority,
            description
        )

        parameters = UrlParameters(entity_id=self.entity.id_)

        request = ApiRequest(
            path=self._PATH,
            method=HTTPMethod.PUT,
            data=DataPackage.from_object(arguments),
            url_parameters=parameters,
            credentials=self.session
        )

        return CustomUnit._decode(
            self.session,
            self.entity,
            request.response_data
        )

    def delete(
        self,
        custom_unit_replacement: T = None,
        global_unit_replacement: GlobalUnit = None
    ) -> None:
        """
        Irrecoverably delete this Custom Unit. Supply either a Custom Unit
        or Global Unit with which to replace any instances of this
        Custom Unit presently denominating an Account. Even if you are
        certain that this Custom Unit does not presently denominated any
        Account, you must supply a replacement unit.
        """
        raise NotImplementedError

    class CreationArguments(Encodable):
        """
        Used by instances of class CustomUnit to validate arguments provided
        for the creation of a new Custom Unit
        """
        def __init__(
            self,
            code: str,
            name: str,
            exponent: int,
            priority: Optional[int] = None,
            description: Optional[str] = None
        ) -> None:

            super().__init__()

            self._code = CustomUnit._Code(code)
            self._name = CustomUnit._Name(name)
            self._priority = CustomUnit._Priority(priority)
            self._description = CustomUnit._Description(description)
            self._exponent = CustomUnit._Exponent(exponent)

            return

        def serialise(self) -> dict:
            data = {
                'code': self._code.serialise(),
                'name': self._name.serialise(),
                'priority': self._priority.serialise(),
                'description': self._description.serialise(),
                'exponent': self._exponent.serialise()
            }
            return data

    class UpdateArguments(Encodable):
        def __init__(
            self,
            custom_unit_id: int,
            code: str,
            name: str,
            exponent: int,
            priority: Optional[int] = None,
            description: Optional[str] = None
        ) -> None:

            if not isinstance(custom_unit_id, int):
                raise TypeError('custom_unit_id must be of type `int`')

            self._id = custom_unit_id

            self._code = CustomUnit._Code(code)
            self._name = CustomUnit._Name(name)
            self._priority = CustomUnit._Priority(priority)
            self._description = CustomUnit._Description(description)
            self._exponent = CustomUnit._Exponent(exponent)

            return

        def serialise(self) -> Dict[str, Any]:
            data = {
                'custom_unit_id': self._id,
                'code': self._code.serialise(),
                'name': self._name.serialise(),
                'priority': self._priority.serialise(),
                'description': self._description.serialise(),
                'exponent': self._exponent.serialise()
            }
            return data

    class _Priority(Encodable):
        def __init__(self, priority: Optional[int]) -> None:

            self._priority = priority
            if priority is None:
                return

            self._priority = ConstrainedInteger(
                priority,
                'priority',
                CustomUnit.MAX_PRIORITY_VALUE,
                CustomUnit.MIN_PRIORITY_VALUE
            )
            return

        def serialise(self) -> Optional[int]:
            if self._priority is None:
                return None
            assert isinstance(self._priority, ConstrainedInteger)
            return self._priority.serialise()

    class _Code(ConstrainedString):
        def __init__(self, code: str) -> None:
            super().__init__(
                code,
                'code',
                CustomUnit.MAX_CODE_LENGTH,
                CustomUnit.MIN_CODE_LENGTH
            )
            return

    class _Name(ConstrainedString):
        def __init__(self, name: str) -> None:
            super().__init__(
                name,
                'name',
                CustomUnit.MAX_NAME_LENGTH
            )
            return

    class _Description(ConstrainedString):
        def __init__(self, description: Optional[str]) -> None:
            if description is None:
                description = ''
            super().__init__(
                description,
                'description',
                CustomUnit.MAX_DESCRIPTION_LENGTH
            )
            return

    class _Exponent(ConstrainedInteger):
        def __init__(self, exponent: int) -> None:
            super().__init__(
                exponent,
                'exponent',
                CustomUnit.MAX_EXPONENT_VALUE,
                CustomUnit.MIN_EXPONENT_VALUE
            )
            return

    def __repr__(self) -> str:
        rep = '<amatino.CustomUnit at {memory}, id: {id_}, code: {code}, '
        rep += 'name: {name}, priority: {priority}, exponent: {exponent}, '
        rep += 'description: {description}, session: {session_id}, '
        rep += 'entity: {entity_id}>'
        representation = rep.format(
            memory=hex(id(self)),
            id_=str(self.id_),
            code=self.code,
            name=self.name,
            priority=self.priority,
            exponent=self.exponent,
            description=self.description,
            session_id=str(self.session.id_),
            entity_id=self.entity.id_
        )
        return representation



"""
Amatino API Python Bindings
Ledger Module
Author: hugh@amatino.io
"""
from datetime import datetime
from amatino.ledger_order import LedgerOrder
from amatino.internal.am_time import AmatinoTime
from amatino.denomination import Denomination
from amatino.custom_unit import CustomUnit
from amatino.global_unit import GlobalUnit
from amatino.session import Session
from amatino.entity import Entity
from amatino.internal.immutable import Immutable
from amatino.account import Account
from amatino.internal.encodable import Encodable
from amatino.internal.api_request import ApiRequest
from amatino.internal.url_parameters import UrlParameters
from amatino.ledger_row import LedgerRow
from amatino.unexpected_response_type import UnexpectedResponseType
from amatino.missing_key import MissingKey
from amatino.internal.http_method import HTTPMethod
from amatino.internal.data_package import DataPackage
from amatino.internal.am_amount import AmatinoAmount
from typing import Optional
from typing import TypeVar
from typing import Type
from typing import Dict
from typing import Any
from typing import List

T = TypeVar('T', bound='Ledger')


class Ledger:
    """
    A Ledger is a list of Transactions from the perspective of a particular
    Account. Ledgers are ordered by Transaction time, and include a running
    Account Balance for every line.

    You can request Ledgers in arbitrary Global or Custom Units, not just the
    native unit of the target Account. If you request a Ledger in a unit other
    than the target Account native unit, Amatino will compute and return
    unrealised gains and losses.

    Amatino will return a maximum total of 1,000 Ledger Rows per retrieval
    request. If the Ledger you define spans more than 1,000 rows, it will be
    broken into pages you can retrieve seperately.
    """

    _PATH = '/accounts/ledger'

    def __init__(
        self,
        session: Session,
        entity: Entity,
        account_id: int,
        start_time: AmatinoTime,
        end_time: AmatinoTime,
        recursive: bool,
        generated_time: AmatinoTime,
        global_unit_id: Optional[int],
        custom_unit_id: Optional[int],
        page: int,
        number_of_pages: int,
        order: LedgerOrder,
        ledger_rows: List[LedgerRow]
    ) -> None:

        self._session = session
        self._entity = entity
        self._account_id = account_id
        self._start_time = start_time
        self._end_time = end_time
        self._recursive = recursive
        self._generated_time = generated_time
        self._global_unit_id = global_unit_id
        self._custom_unit_id = custom_unit_id
        self._page = page
        self._number_of_pages = number_of_pages
        self._order = order
        self._rows = ledger_rows

        return

    session = Immutable(lambda s: s._session)
    entity = Immutable(lambda s: s._entity)
    account_id = Immutable(lambda s: s._account_id)
    account = Immutable(
        lambda s: Account.retrieve(s.session, s.entity, s.account_id)
    )
    denomination = Immutable(lambda s: s._denomination())
    start_time = Immutable(lambda s: s._start_time.raw)
    end_time = Immutable(lambda s: s._end_time.raw)
    recursive = Immutable(lambda s: s._recursive)
    generated_time = Immutable(lambda s: s._generated_time.raw)
    global_unit_id = Immutable(lambda s: s._global_unit_id)
    custom_unit_id = Immutable(lambda s: s._custom_unit_id)
    page = Immutable(lambda s: s._page)
    number_of_pages = Immutable(lambda s: s._number_of_pages)
    order = Immutable(lambda s: s._order)
    rows = Immutable(lambda s: s._rows)

    def _denomination(self) -> Denomination:
        """Return the underlying denomination of this Ledger"""
        if self._global_unit_id is None:
            return CustomUnit.retrieve(
                self.entity,
                self.session,
                self._custom_unit_id
            )
        return GlobalUnit.retrieve(self.session, self._global_unit_id)

    @classmethod
    def retrieve(
        cls: Type[T],
        session: Session,
        entity: Entity,
        account: Account,
        order: LedgerOrder = LedgerOrder.YOUNGEST_FIRST,
        page: int = 1,
        start_time: Optional[datetime] = None,
        end_time: Optional[datetime] = None,
        denomination: Optional[Denomination] = None
    ) -> T:
        """
        Retrieve a Ledger for the supplied account. Optionally specify order,
        page, denomination, start time, and end time.
        """
        if not isinstance(session, Session):
            raise TypeError('session must be of type `Session`')

        if not isinstance(entity, Entity):
            raise TypeError('entity must be of type `Entity`')

        arguments = Ledger.RetrieveArguments(
            account,
            order,
            page,
            start_time,
            end_time,
            denomination
        )
        data = DataPackage(object_data=arguments, override_listing=True)

        parameters = UrlParameters(entity_id=entity.id_)

        request = ApiRequest(
            path=cls._PATH,
            method=HTTPMethod.GET,
            credentials=session,
            data=data,
            url_parameters=parameters
        )

        return cls._decode(session, entity, request.response_data)

    @classmethod
    def _decode(
        cls: Type[T],
        session: Session,
        entity: Entity,
        data: Any
    ) -> T:

        if not isinstance(data, dict):
            raise UnexpectedResponseType(type(data), dict)

        try:
            ledger = cls(
                session,
                entity,
                account_id=data['account_id'],
                start_time=AmatinoTime.decode(data['start_time']),
                end_time=AmatinoTime.decode(data['end_time']),
                recursive=data['recursive'],
                generated_time=AmatinoTime.decode(data['generated_time']),
                global_unit_id=data['global_unit_denomination'],
                custom_unit_id=data['custom_unit_denomination'],
                ledger_rows=Ledger._decode_rows(data['ledger_rows']),
                page=data['page'],
                number_of_pages=data['number_of_pages'],
                order=LedgerOrder(data['ordered_oldest_first'])
            )
        except KeyError as error:
            raise MissingKey(error.args[0])

        return ledger

    @classmethod
    def _decode_rows(cls: Type[T], rows: List[Any]) -> List[LedgerRow]:
        """Return LedgerRows decoded from raw API response data"""
        if not isinstance(rows, list):
            raise UnexpectedResponseType(rows, list)

        def decode(data) -> LedgerRow:

            if not isinstance(data, list):
                raise UnexpectedResponseType(data, list)

            row = LedgerRow(
                transaction_id=data[0],
                transaction_time=AmatinoTime.decode(data[1]),
                description=data[2],
                opposing_account_id=data[3],
                opposing_account_name=data[4],
                debit=AmatinoAmount.decode(data[5]),
                credit=AmatinoAmount.decode(data[6]),
                balance=AmatinoAmount.decode(data[7])
            )

            return row

        return [decode(r) for r in rows]

    class RetrieveArguments(Encodable):
        def __init__(
            self,
            account: Account,
            order: LedgerOrder = LedgerOrder.YOUNGEST_FIRST,
            page: int = 1,
            start_time: Optional[datetime] = None,
            end_time: Optional[datetime] = None,
            denomination: Optional[Denomination] = None
        ) -> None:

            if not isinstance(account, Account):
                raise TypeError('account must be of type `Account`')

            if not isinstance(order, LedgerOrder):
                raise TypeError('order must beof type `LedgerOrder`')

            if not isinstance(page, int):
                raise TypeError('page must be of type `int`')

            if start_time and not isinstance(start_time, datetime):
                raise TypeError('start_time must be of type `datetime` or None')

            if end_time and not isinstance(end_time, datetime):
                raise TypeError('end_time must be of type `datetime` or None')

            if denomination and not isinstance(denomination, Denomination):
                raise TypeError(
                    'denomination must be of type `Denomination` or None'
                )

            if denomination is None:
                denomination = account.denomination

            self._account = account
            self._order = order
            self._page = page
            self._start_time = None
            if start_time:
                self._start_time = AmatinoTime(start_time)
            self._end_time = None
            if end_time:
                self._end_time = AmatinoTime(end_time)
            self._denomination = denomination

        def serialise(self) -> Dict[str, Any]:
            global_unit_id = None
            custom_unit_id = None
            if isinstance(self._denomination, GlobalUnit):
                global_unit_id = self._denomination.id_
            else:
                assert isinstance(self._denomination, CustomUnit)
                custom_unit_id = self._denomination.id_

            start_time = None
            if self._start_time:
                start_time = AmatinoTime(self._start_time).serialise()

            end_time = None
            if self._end_time:
                end_time = AmatinoTime(self._end_time).serialise()

            data = {
                'account_id': self._account.id_,
                'start_time': start_time,
                'end_time': end_time,
                'page': self._page,
                'global_unit_denomination': global_unit_id,
                'custom_unit_denomination': custom_unit_id,
                'order_oldest_first': self._order.value
            }

            return data

    def __iter__(self):
        return Ledger.Iterator(self._rows)

    class Iterator:
        """An iterator for iterating through LedgerRows in a Ledger"""

        def __init__(self, rows: List[LedgerRow]) -> None:
            self._index = 0
            self._rows = rows
            return

        def __next__(self) -> LedgerRow:
            if self._index >= len(self._rows):
                raise StopIteration
            row = self._rows[self._index]
            self._index += 1
            return row

    def __len__(self):
        return len(self.rows)



"""
Amatino API Python Bindings
Global Unit Module
Author: hugh@amatino.io
"""
from amatino.session import Session
from amatino.denomination import Denomination
from amatino.internal.url_target import UrlTarget
from amatino.internal.url_parameters import UrlParameters
from amatino.internal.api_request import ApiRequest
from amatino.internal.immutable import Immutable
from amatino.internal.http_method import HTTPMethod
from amatino.api_error import ApiError
from amatino.unexpected_response_type import UnexpectedResponseType
from typing import TypeVar
from typing import Type
from typing import Any
from typing import List

T = TypeVar('T', bound='GlobalUnit')


class GlobalUnit(Denomination):
    """
    Global Units are standardised units of account available across
    all Amatino Entities. For example, many major currencies are available
    as Global Units.

    Global Units cannot be modified by Amatino users.
    """
    _PATH = '/units'
    _URL_KEY = 'global_unit_id'

    def __init__(
        self,
        session: Session,
        code: str,
        id_: int,
        name: str,
        priority: int,
        description: str,
        exponent: int
    ) -> None:

        assert isinstance(session, Session)
        self._session = session
        super().__init__(code, id_, name, priority, description, exponent)

        return

    session = Immutable(lambda s: s._session)

    @classmethod
    def retrieve(cls: Type[T], session: Session, id_: int) -> T:
        """Retrieve a Global Unit"""
        if not isinstance(id_, int):
            raise TypeError('id_ must be of type `int`')
        return GlobalUnit.retrieve_many(session, [id_])[0]

    @classmethod
    def retrieve_many(
        cls: Type[T],
        session: Session,
        ids: List[int]
    ) -> List[T]:
        """Retrieve a set of Global Units"""

        if not isinstance(session, Session):
            raise TypeError('session must be of type `Session`')

        if not isinstance(ids, list):
            raise TypeError('ids must be of type `List[int]`')

        if False in [isinstance(i, int) for i in ids]:
            raise TypeError('ids must be of type `List[int]`')

        targets = UrlTarget.from_many_integers(GlobalUnit._URL_KEY, ids)
        parameters = UrlParameters.from_targets(targets)

        request = ApiRequest(
            path=GlobalUnit._PATH,
            url_parameters=parameters,
            credentials=session,
            method=HTTPMethod.GET
        )

        units = GlobalUnit._decode_many(session, request.response_data)

        return units

    @classmethod
    def _decode_many(cls: Type[T], session: Session, data: Any) -> List[T]:
        """Return a list of Global Units decoded from raw API response data"""

        if not isinstance(data, list):
            raise ApiError('Unexpected non-list API data response')

        def decode(unit_data: Any) -> T:
            if not isinstance(unit_data, dict):
                raise UnexpectedResponseType(unit_data, dict)

            try:
                unit = cls(
                    session=session,
                    code=unit_data['code'],
                    id_=unit_data['global_unit_id'],
                    name=unit_data['name'],
                    priority=unit_data['priority'],
                    description=unit_data['description'],
                    exponent=unit_data['exponent']
                )
            except KeyError as error:
                message = 'Expected key "{key}" missing from response data'
                message.format(key=error.args[0])
                raise ApiError(message)

            return unit

        units = [decode(u) for u in data]

        return units



"""
Amatino API Python Bindings
User Module
Author: hugh@amatino.io
"""
from amatino.internal.immutable import Immutable
from amatino.session import Session
from amatino.internal.url_parameters import UrlParameters
from amatino.internal.url_target import UrlTarget
from amatino.internal.api_request import ApiRequest
from amatino.internal.http_method import HTTPMethod
from amatino.unexpected_response_type import UnexpectedResponseType
from typing import TypeVar
from typing import Type
from typing import Optional
from typing import List
from typing import Any

T = TypeVar('T', bound='User')


class User:
    """
    A User is a human producer and consumer of data stored by Amatino. When you
    create an Amatino account on the Amatino website, a User is generated in
    your name. You can create other Users at will to serve the needs of your
    application. For example, you might wish to create an Amatino User to
    associate with each end-user of your application, in order to link financial
    information stored in Amatino with that end-user.

    Users created via the Amatino API cannot login or otherwise interact with
    the amatino.io website in any way. They are not eligbile to receive customer
    support from us directly (though you are most welcome to request customer
    support to assist you with users you create), and don't generate associated
    discussion forum accounts. You have absolute control over their lifecycle.
    They can make requests to the Amatino API on their own behalf.

    Generally, if you are creating User accounts for your fellow developers,
    you will want to do so in your billing dashboard. Doing so will allow them
    to manage their password, post to the discussion forums, and contact us for
    support. If you are creating Users to manage financial data inside your
    application, you will want to do so via the Amatino API.

    Users and Entities are woven together using permission graphs. Any User may
    be granted read and or write access to any Account in any Entity, whether
    they were created in the billing dashboard or via the Amatino API.

    If you are on a Fixed Price plan, each additional user you create in the
    Amatino API will count towards your monthly bill. If you are on a Pay Per
    Use plan, creating additional Users incurs no direct marginal cost. You can
    change your plan at any time.
    """
    _URL_KEY = 'user_id'
    _PATH = '/users'

    def __init__(
        self,
        session: Session,
        id_: int,
        email: str,
        name: str,
        handle: str,
        avatar_url: str
    ) -> None:

        self._id = id_
        self._email = email
        self._name = name
        self._handle = handle
        self._avatar_url = avatar_url

        return

    id_ = Immutable(lambda s: s._id)
    email = Immutable(lambda s: s._email)
    name = Immutable(lambda s: s._name)
    handle = Immutable(lambda s: s._handle)
    avatar_url = Immutable(lambda s: s._avatar_url)

    @classmethod
    def retrieve_authenticated_user(cls: Type[T], session: Session) -> T:
        """Return the User authenticated by the supplied Session"""
        raise NotImplementedError  # Pending bug fix in User retrieval

    @classmethod
    def retrieve(cls: Type[T], session: Session, id_: int) -> T:
        """Return the User with supplied ID"""
        if not isinstance(id_, int):
            raise TypeError('id_ must be of type `int`')
        return cls.retrieve_many(session, [id_])[0]

    @classmethod
    def retrieve_many(
        cls: Type[T],
        session: Session,
        ids: List[int]
    ) -> List[T]:
        """Return a list of Users"""
        if not isinstance(ids, list) or False in [
            isinstance(i, int) for i in ids
        ]:
            raise TypeError('ids must be of type List[int]')
        return cls._retrieve_many(session, ids)

    @classmethod
    def _retrieve_many(
        cls: Type[T],
        session: Session,
        ids: Optional[List[int]] = None
    ) -> List[T]:
        """Return a list of users"""

        parameters = None
        if ids is not None:
            targets = UrlTarget.from_many_integers(cls._URL_KEY, ids)
            parameters = UrlParameters.from_targets(targets)

        request = ApiRequest(
            path=cls._PATH,
            data=None,
            credentials=session,
            method=HTTPMethod.GET,
            url_parameters=parameters
        )

        users = cls._decode_many(session, request.response_data)

        return users

    @classmethod
    def _decode_many(cls: Type[T], session: Session, data: Any) -> List[T]:
        """Return a list of Users decoded from API response data"""

        if not isinstance(data, list):
            raise UnexpectedResponseType(data, list)

        def decode(obj: Any) -> T:
            if not isinstance(obj, dict):
                raise UnexpectedResponseType(obj, dict)

            user = cls(
                session=session,
                id_=obj['user_id'],
                email=obj['email'],
                name=obj['name'],
                handle=obj['handle'],
                avatar_url=obj['avatar_url']
            )

            return user

        return [decode(u) for u in data]



"""
Amatino API Python Bindings
Permissions Graph Module
Author: hugh@amatino.io
"""
from amatino.internal.encodable import Encodable
from typing import Optional


class PermissionsGraph(Encodable):
    def __init__(self, raw_graph: Optional[dict]) -> None:
        if raw_graph is None:
            raw_graph = dict()
        self._raw_graph = raw_graph
        return

    def serialise(self) -> dict:
        return self._raw_graph



"""
Amatino API Python Bindings
Balance Test Module
Author: hugh@amatino.io
"""
from amatino.tests.primary.transaction import TransactionTest
from amatino import Balance
from decimal import Decimal

NAME = 'Retrieve a Balance'


class BalanceTest(TransactionTest):
    """Test the Balance object"""

    def __init__(self, name=NAME) -> None:

        super().__init__(name)
        return

    def execute(self) -> None:

        try:
            self.create_transaction(amount=Decimal(42))
            self.create_transaction(amount=Decimal(18))
        except Exception as error:
            self.record_failure(error)
            return

        try:
            balance = Balance.retrieve(self.entity, self.asset)
        except Exception as error:
            self.record_failure(error)
            return

        if balance.magnitude != Decimal(60):
            message = 'Unexpected magnitude: ' + str(balance.magnitude)
            self.record_failure(message)
            return

        self.record_success()
        return



"""
Amatino API Python Bindings
Ledger Test Module
Author: hugh@amatino.io
"""
from amatino.tests.primary.transaction import TransactionTest
from amatino import Ledger
from decimal import Decimal
from amatino import LedgerRow

NAME = 'Retrieve a Ledger'


class LedgerTest(TransactionTest):
    """Test the Ledger object"""

    def __init__(self, name=NAME) -> None:

        super().__init__(name)
        return

    def execute(self) -> None:

        try:
            self.create_transaction(amount=Decimal(42))
            self.create_transaction(amount=Decimal(12))
            self.create_transaction(amount=Decimal(1492))
        except Exception as error:
            self.record_failure(error)
            return

        try:
            ledger = Ledger.retrieve(self.session, self.entity, self.asset)
        except Exception as error:
            self.record_failure(error)
            return

        for row in ledger:
            if not isinstance(row, LedgerRow):
                self.record_failure('Unexpected non-LedgerRow type')
                return

        if len(ledger) != 3:
            self.record_failure('Unexpected number of ledger rows')
            return

        self.record_success()



"""
Amatino API Python Bindings
Constrained String Module
Author: hugh@amatino.io

This module is intended to be private, used indirectly by public classes, and
should not be used directly.

"""
from typing import Optional
from amatino.constraint_error import ConstraintError
from amatino.internal.encodable import Encodable


class ConstrainedString(Encodable):
    """
    A string whose maximum, and optionally minimum, length is restricted. Throws
    TypeError if supplied with something other than a `str`, and ConstraintError
    if a constrained is violated.
    """

    MAX_ERR = "{name} exceeds maximum length of {max_char}"
    MIN_ERR = "{name} below minimum length of {min_char}"

    def __init__(
        self,
        string: str,
        name: str,
        max_length: int,
        min_length: Optional[int] = None
    ) -> None:

        assert isinstance(name, str)
        assert isinstance(max_length, int)
        if min_length is not None:
            assert isinstance(min_length, int)

        if not isinstance(string, str):
            raise TypeError(name + ' must be of type `str`')

        if len(string) > max_length:
            error = self.MAX_ERR.format(name=name, max_char=str(max_length))
            raise ConstraintError(error)

        self._string = string
        self._name = name

        if min_length is None:
            return

        if len(string) < min_length:
            error = self.MIN_ERR.format(name=name, min_char=str(min_length))
            raise ConstraintError(error)

        return

    def __str__(self) -> str:
        return self._string

    def serialise(self) -> str:
        return self._string



"""
Amatino API Python Bindings
Balance Protocol Module
Author: hugh@blinkybeach.com
"""
from typing import TypeVar
from typing import Type
from typing import List
from typing import Dict
from typing import Any
from typing import Optional
from decimal import Decimal
from datetime import datetime
from amatino.denomination import Denomination
from amatino.entity import Entity
from amatino.account import Account
from amatino.global_unit import GlobalUnit
from amatino.custom_unit import CustomUnit
from amatino.unexpected_response_type import UnexpectedResponseType
from amatino.internal.am_time import AmatinoTime
from amatino.internal.immutable import Immutable
from amatino.internal.encodable import Encodable
from amatino.internal.api_request import ApiRequest
from amatino.internal.http_method import HTTPMethod
from amatino.internal.data_package import DataPackage
from amatino.internal.url_parameters import UrlParameters


T = TypeVar('T', bound='BalanceProtocol')
K = TypeVar('K', bound='BalanceProtocol.RetrieveArguments')


class BalanceProtocol:
    """
    Abstract class defining a protocol for classes representing balances. In
    practice these are Balance and RecrusiveBalance. This class is intended
    to be private, and internal to the Amatino library. You should not
    use it directly when integrating Amatino Python into your application.
    """

    PATH = NotImplemented

    def __init__(
        self,
        entity: Entity,
        balance_time: AmatinoTime,
        generated_time: AmatinoTime,
        recursive: bool,
        global_unit_id: Optional[int],
        custom_unit_id: Optional[int],
        account_id: int,
        magnitude: Decimal
    ) -> None:

        if self.PATH == NotImplemented:
            raise RuntimeError('Balance classes must implement .PATH property')

        assert isinstance(entity, Entity)
        assert isinstance(balance_time, AmatinoTime)
        assert isinstance(generated_time, AmatinoTime)
        assert isinstance(recursive, bool)
        if global_unit_id is not None:
            assert isinstance(global_unit_id, int)
        if custom_unit_id is not None:
            assert isinstance(custom_unit_id, int)
        assert isinstance(account_id, int)
        assert isinstance(magnitude, Decimal)

        self._entity = entity
        self._balance_time = balance_time
        self._generated_time = generated_time
        self._recursive = recursive
        self._custom_unit_id = custom_unit_id
        self._global_unit_id = global_unit_id
        self._magnitude = magnitude
        self._account_id = account_id

        return

    entity = Immutable(lambda s: s._entity)
    session = Immutable(lambda s: s._entity.session)
    time = Immutable(lambda s: s._balance_time.raw)
    generated_time = Immutable(lambda s: s._generated_time.raw)
    magnitude = Immutable(lambda s: s._magnitude)
    is_recursive = Immutable(lambda s: s._recursive)
    account_id = Immutable(lambda s: s._account_id)
    account = Immutable(
        lambda s: Account.retrieve(s.entity.session, s.entity, s.account_id)
    )
    denomination = Immutable(lambda s: s._denomination())
    global_unit_id = Immutable(lambda s: s._global_unit_id)
    custom_unit_id = Immutable(lambda s: s._custom_unit_id)

    def _denomination(self) -> Denomination:
        """Return the underlying denomination of this Balance"""
        if self._global_unit_id is None:
            return CustomUnit.retrieve(
                self.entity,
                self.session,
                self._custom_unit_id
            )
        return GlobalUnit.retrieve(self.session, self._global_unit_id)

    @classmethod
    def retrieve_many(
        cls: Type[T],
        entity: Entity,
        arguments: List[K]
    ) -> List[T]:
        """Retrieve several Balances."""
        if not isinstance(entity, Entity):
            raise TypeError('entity must be of type `Entity`')

        if False in [isinstance(a, cls.RetrieveArguments) for a in arguments]:
            raise TypeError(
                'arguments must be of type List[Balance.RetrieveArguments]'
            )

        data = DataPackage(list_data=arguments)
        parameters = UrlParameters(entity_id=entity.id_)

        request = ApiRequest(
            path=cls.PATH,
            method=HTTPMethod.GET,
            credentials=entity.session,
            data=data,
            url_parameters=parameters
        )

        return cls._decode_many(entity, request.response_data)

    @classmethod
    def _decode_many(
        cls: Type[T],
        entity: Entity,
        data: Any
    ) -> List[T]:

        if not isinstance(data, list):
            raise UnexpectedResponseType(data, list)

        for balance in data:
            if not isinstance(balance, dict):
                raise UnexpectedResponseType(balance, dict)

        balances = list()

        for balance in data:
            balances.append(cls(
                entity,
                AmatinoTime.decode(balance['balance_time']),
                AmatinoTime.decode(balance['generated_time']),
                balance['recursive'],
                balance['global_unit_denomination'],
                balance['custom_unit_denomination'],
                balance['account_id'],
                Decimal(balance['balance'])
            ))

        return balances

    @classmethod
    def retrieve(
        cls: Type[T],
        entity: Entity,
        account: Account,
        balance_time: Optional[datetime] = None,
        denomination: Optional[Denomination] = None
    ) -> T:
        """Retrieve a Balance"""
        arguments = cls.RetrieveArguments(
            account,
            balance_time,
            denomination
        )
        return cls.retrieve_many(entity, [arguments])[0]

    class RetrieveArguments(Encodable):
        def __init__(
            self,
            account: Account,
            balance_time: Optional[datetime] = None,
            denomination: Optional[Denomination] = None
        ) -> None:

            if not isinstance(account, Account):
                raise TypeError('account must be of type Account')

            if (
                    balance_time is not None
                    and not isinstance(balance_time, datetime)
            ):
                raise TypeError(
                    'balance_time must be of type `datetime or None'
                )

            if (
                    denomination is not None
                    and not isinstance(denomination, Denomination)
            ):
                raise TypeError(
                    'denomination must conform to `Denomination`'
                )

            if denomination is None:
                denomination = account.denomination

            self._account = account
            self._balance_time = None
            if balance_time:
                self._balance_time = AmatinoTime(balance_time)
            self._denomination = denomination

        def serialise(self) -> Dict[str, Any]:

            global_unit_id = None
            custom_unit_id = None

            if isinstance(self._denomination, GlobalUnit):
                global_unit_id = self._denomination.id_
            else:
                assert isinstance(self._denomination, CustomUnit)
                custom_unit_id = self._denomination.id_

            balance_time = None
            if self._balance_time:
                balance_time = self._balance_time.serialise()

            data = {
                'account_id': self._account.id_,
                'balance_time': balance_time,
                'custom_unit_denomination': custom_unit_id,
                'global_unit_denomination': global_unit_id
            }

            return data



from amatino.tests.ancillary.session import SessionTest
from amatino.tests.ancillary.global_unit import GlobalUnitTest
from amatino.tests.ancillary.user import UserTest
#from amatino.tests.ancillary.custom_unit import CustomUnitTest



"""
Amatino API Python Bindings
New Transaction Arguments Module
Author: hugh@amatino.io

This module is intended to be private, used indirectly by public classes, and
should not be used directly.
"""

from amatino.internal.encodable import Encodable


class NewSessionArguments(Encodable):
    """
    Private - Not intended to be used directly.

    An instance of arguments for the creation of a new Session. Performs
    validation of input parameters. May raise various exceptions during
    validation.
    """
    _INVALID_INIT_ID = """Supply either email or user_id, but not both"""

    def __init__(
        self,
        secret: str = None,
        email: str = None,
        user_id: int = None
    ) -> None:

        if email is not None and user_id is not None:
            raise ValueError(self._INVALID_INIT_ID)

        if not isinstance(secret, str):
            raise TypeError('secret must be of type str')

        if email is not None and not isinstance(email, str):
            raise TypeError('email must be of type str')

        if user_id is not None and not isinstance(user_id, int):
            raise TypeError('user_id must be of type int')

        self._secret = secret
        self._email = email
        self._user_id = user_id

        self._data = {
            'secret': self._secret,
            'account_email': self._email,
            'user_id': self._user_id
        }

        return

    def serialise(self) -> dict:
        return self._data



"""
Amatino API Python Bindings
User Test Module
Author: hugh@amatino.io
"""
from amatino import User
from amatino.tests.ancillary.session import SessionTest


class UserTest(SessionTest):
    """Test the User ancillary object"""

    def __init__(self, name='Retrieve a User') -> None:

        super().__init__(name)
        self.create_session()
        return

    def execute(self) -> None:

        try:
            user = User.retrieve(self.session, self.user_id)
        except Exception as error:
            self.record_failure(error)
            return

        if not isinstance(user, User):
            return_type = str(type(user))
            self.record_failure('Unexpected return type: ' + return_type)
            return

        if not isinstance(user.id_, int):
            id_type = str(type(user.id_))
            self.record_failure('Unexpected user ID type: ' + id_type)
            return

        self.record_success()
        return



"""
Amatino API Python Bindings
Performance class Test Module
Author: hugh@blinkybeach.com
"""
from amatino.tests.primary.account import AccountTest
from amatino import Account
from amatino import Transaction
from amatino import Performance
from amatino import AMType
from amatino import Side
from amatino import Entry
from amatino import GlobalUnit
from amatino import TreeNode
from amatino import Entity
from datetime import datetime
from datetime import timedelta
from decimal import Decimal

NAME = 'Retrieve a Performance'


class PerformanceTest(AccountTest):
    """Test the Performance object"""

    def __init__(self, name=NAME) -> None:
        super().__init__(name)
        return

    def execute(self) -> None:

        try:

            income = self.create_account(
                amt=AMType.income,
                name='Test Income'
            )

            income_child = Account.create(
                self.session,
                self.entity,
                'Test income child',
                income.am_type,
                income.denomination,
                parent=income
            )

            expense = self.create_account(
                amt=AMType.expense,
                name='Test Expense'
            )

            Transaction.create(
                self.session,
                self.entity,
                datetime.utcnow(),
                [
                    Entry(Side.credit, Decimal('400'), income),
                    Entry(Side.debit, Decimal('400'), expense)
                ],
                self.usd,
                'Test transaction 1'
            )

            Transaction.create(
                self.session,
                self.entity,
                datetime.utcnow(),
                [
                    Entry(Side.credit, Decimal('250'), income_child),
                    Entry(Side.credit, Decimal('250'), income),
                    Entry(Side.debit, Decimal('500'), expense)
                ],
                self.usd,
                'Test transaction 2'
            )

            Transaction.create(
                self.session,
                self.entity,
                datetime.utcnow() - timedelta(days=4),
                [
                    Entry(Side.credit, Decimal('1'), income),
                    Entry(Side.debit, Decimal('1'), expense)
                ],
                self.usd,
                'Test transaction 3'
            )

            Transaction.create(
                self.session,
                self.entity,
                datetime.utcnow() + timedelta(days=4),
                [
                    Entry(Side.debit, Decimal('50'), income),
                    Entry(Side.credit, Decimal('50'), expense)
                ],
                self.usd,
                'Test transaction 4'
            )

            performance = Performance.retrieve(
                entity=self.entity,
                start_time=datetime.utcnow() - timedelta(days=2),
                end_time=datetime.utcnow() + timedelta(days=2),
                denomination=self.usd
            )

            assert isinstance(performance, Performance)
            assert performance.has_income
            assert performance.has_expenses
            assert performance.global_unit_id == self.usd.id_
            assert performance.custom_unit_id is None
            assert isinstance(performance.denomination, GlobalUnit)
            assert isinstance(performance.start_time, datetime)
            assert isinstance(performance.end_time, datetime)
            assert isinstance(performance.generated_time, datetime)
            assert isinstance(performance.entity, Entity)

            income_root = performance.income[0]
            expense_root = performance.expenses[0]

            assert isinstance(income_root, TreeNode)
            assert isinstance(expense_root, TreeNode)
            assert income_root.has_children
            assert not expense_root.has_children
            assert isinstance(income_root.children[0], TreeNode)

            assert income_root.recursive_balance == Decimal('900')
            assert income_root.account_balance == Decimal('650')
            assert income_root.children[0].account_balance == Decimal('250')
            assert expense_root.account_balance == Decimal('900')
            assert expense_root.recursive_balance == Decimal('900')

        except Exception as error:
            self.record_failure(error)
            return

        self.record_success()



"""
Amatino API Python Bindings
Url Target Module
Author: hugh@amatino.io
"""
from typing import Type
from typing import TypeVar
from typing import List

T = TypeVar('T', bound='UrlTarget')


class UrlTarget:
    """Internal class handling individual url parameter components"""

    def __init__(self, key: str, value: str) -> None:

        if not isinstance(key, str):
            raise TypeError('Key must be of type `str`')

        if not isinstance(value, str):
            raise TypeError('value must be of type `str`')

        self.key = key
        self.value = value

        return

    @classmethod
    def from_integer(cls: Type[T], key: str, value: int) -> T:
        """Return a UrlTarget formed of an integer value"""
        return cls(key, str(value))

    @classmethod
    def from_many_integers(
        cls: Type[T],
        key: str,
        values: List[int]
    ) -> List[T]:
        """Return a list of integer targets sharing a key"""
        return [cls(key, str(v)) for v in values]

    def __str__(self):
        return self.key + '=' + str(self.value)

    def __eq__(self, other):
        if (
                other.key == self.key
                and other.value == self.value
        ):
            return True
        return False



"""
Amatino API Python Bindings
Transaction Module
Author: hugh@amatino.io
"""
from datetime import datetime
from amatino.session import Session
from amatino.entity import Entity
from amatino.global_unit import GlobalUnit
from amatino.custom_unit import CustomUnit
from amatino.entry import Entry
from amatino.side import Side
from amatino.denomination import Denomination
from amatino.internal.encodable import Encodable
from amatino.internal.constrained_string import ConstrainedString
from amatino.internal.am_time import AmatinoTime
from amatino.internal.api_request import ApiRequest
from amatino.internal.data_package import DataPackage
from amatino.internal.http_method import HTTPMethod
from amatino.internal.url_target import UrlTarget
from amatino.internal.url_parameters import UrlParameters
from amatino.api_error import ApiError
from amatino.missing_key import MissingKey
from amatino.internal.am_amount import AmatinoAmount
from decimal import Decimal
from typing import TypeVar
from typing import Optional
from typing import Type
from typing import Any
from typing import List
from typing import Dict
from amatino.internal.immutable import Immutable

T = TypeVar('T', bound='Transaction')


class Transaction:
    """
    A Transaction is an exchange of value between two or more Accounts. For
    example, the raising of an invoice, the incrurring of a liability, or the
    receipt of a payment. Many Transactions together, each touching the same
    Account, form an Account Ledger, and the cumulative sum of the Transactions
    form an Account Balance.

    Transactions are composed of Entries, each of which includes a debit or
    credit (The fundamental Sides). The sum of all debits and credits in all
    the Entries that compose a Transaction must always equal zero.

    Transactions may be retrieved and created in arbitrary units, either a
    Global Units or Custom Units. Amatino will transparently handle all unit
    conversions. For example, a Transaction could be created in Australian
    Dollars, touch an Account denominated in Pounds Sterling, and be retrieved
    in Bitcoin.
    """
    _PATH = '/transactions'
    MAX_DESCRIPTION_LENGTH = 1024
    _URL_KEY = 'transaction_id'

    def __init__(
        self,
        session: Session,
        entity: Entity,
        transaction_id: int,
        transaction_time: AmatinoTime,
        description: str,
        entries: List[Entry],
        global_unit_id: Optional[int] = None,
        custom_unit_id: Optional[int] = None
    ) -> None:

        self._session = session
        self._entity = entity
        self._id = transaction_id
        self._time = transaction_time
        self._description = description
        self._entries = entries
        self._global_unit_id = global_unit_id
        self._custom_unit_id = custom_unit_id

        return

    session = Immutable(lambda s: s._session)
    entity = Immutable(lambda s: s._entity)
    id_ = Immutable(lambda s: s._id)
    time = Immutable(lambda s: s._time.raw)
    description = Immutable(lambda s: s._description)
    entries = Immutable(lambda s: s._entries)
    global_unit_id = Immutable(lambda s: s._global_unit_id)
    custom_unit_id = Immutable(lambda s: s._custom_unit_id)
    denomination = Immutable(lambda s: s._denomination())

    @classmethod
    def create(
        cls: Type[T],
        session: Session,
        entity: Entity,
        time: datetime,
        entries: List[Entry],
        denomination: Denomination,
        description: Optional[str] = None,
    ) -> T:

        arguments = Transaction.CreateArguments(
            time,
            entries,
            denomination,
            description
        )

        data = DataPackage.from_object(arguments)
        parameters = UrlParameters(entity_id=entity.id_)

        request = ApiRequest(
            path=Transaction._PATH,
            method=HTTPMethod.POST,
            credentials=session,
            data=data,
            url_parameters=parameters
        )

        transaction = cls._decode(session, entity, request.response_data)

        return transaction

    @classmethod
    def retrieve(
        cls: Type[T],
        session: Session,
        entity: Entity,
        id_: int,
        denomination: Denomination
    ) -> T:
        """Return a retrieved Transaction"""
        return cls.retrieve_many(session, entity, [id_], denomination)[0]

    @classmethod
    def retrieve_many(
        cls: Type[T],
        session: Session,
        entity: Entity,
        ids: List[int],
        denomination: Denomination
    ) -> List[T]:
        """Return many retrieved Transactions"""

        if not isinstance(session, Session):
            raise TypeError('session must be of type `Session`')

        if not isinstance(entity, Entity):
            raise TypeError('entity must be of type `Entity`')

        if not isinstance(ids, list):
            raise TypeError('ids must be of type `list`')

        if False in [isinstance(i, int) for i in ids]:
            raise TypeError('ids must be of type `int`')

        parameters = UrlParameters(entity_id=entity.id_)

        data = DataPackage(list_data=[cls.RetrieveArguments(
            i, denomination, None
        ) for i in ids])

        request = ApiRequest(
            path=Transaction._PATH,
            method=HTTPMethod.GET,
            credentials=session,
            data=data,
            url_parameters=parameters
        )

        transactions = cls._decode_many(
            session,
            entity,
            request.response_data
        )

        return transactions

    @classmethod
    def _decode(
        cls: Type[T],
        session: Session,
        entity: Entity,
        data: List[dict]
    ) -> T:

        return cls._decode_many(session, entity, data)[0]

    @classmethod
    def _decode_many(
        cls: Type[T],
        session: Session,
        entity: Entity,
        data: List[dict]
    ) -> List[T]:

        if not isinstance(data, list):
            raise ApiError('Unexpected non-list data returned')

        if len(data) < 1:
            raise ApiError('Unexpected empty response data')

        def decode(data: dict) -> T:
            if not isinstance(data, dict):
                raise ApiError('Unexpected non-dict data returned')
            try:
                transaction = cls(
                    session=session,
                    entity=entity,
                    transaction_id=data['transaction_id'],
                    transaction_time=AmatinoTime.decode(
                        data['transaction_time']
                    ),
                    description=data['description'],
                    entries=cls._decode_entries(data['entries']),
                    global_unit_id=data['global_unit_denomination'],
                    custom_unit_id=data['custom_unit_denomination']
                )
            except KeyError as error:
                raise MissingKey(error.args[0])

            return transaction

        transactions = [decode(t) for t in data]

        return transactions

    def update(
        self: T,
        time: Optional[datetime] = None,
        entries: Optional[List[Entry]] = None,
        denomination: Optional[Denomination] = None,
        description: Optional[str] = None
    ) -> 'Transaction':
        """Replace existing transaction data with supplied data."""

        arguments = Transaction.UpdateArguments(
            self,
            time,
            entries,
            denomination,
            description
        )

        data = DataPackage.from_object(arguments)
        parameters = UrlParameters(entity_id=self.entity.id_)

        request = ApiRequest(
            path=Transaction._PATH,
            method=HTTPMethod.PUT,
            credentials=self.session,
            data=data,
            url_parameters=parameters
        )

        transaction = Transaction._decode(
            self.session,
            self.entity,
            request.response_data
        )

        if transaction.id_ != self.id_:
            raise ApiError('Mismatched response Trasaction ID - Fatal')

        return transaction

    def delete(self) -> None:
        """
        Destroy this Transaction, such that it will no longer be included
        in any view of this Entity's accounting information.
        """

        target = UrlTarget(self._URL_KEY, str(self.id_))
        parameters = UrlParameters(entity_id=self.entity.id_, targets=[target])

        ApiRequest(
            path=self._PATH,
            method=HTTPMethod.DELETE,
            credentials=self.session,
            data=None,
            url_parameters=parameters
        )

        return

    def restore(self) -> None:
        """
        Restore this transaction from a deleted state to an active state.
        """
        raise NotImplementedError

    def list_versions(self) -> List[Any]:
        """Return a list of versions of this Transaction"""
        raise NotImplementedError

    def _denomination(self) -> Denomination:
        """Return the Denomination of this Transaction"""
        if self.global_unit_id is not None:
            return GlobalUnit.retrieve(self.session, self.global_unit_id)
        return CustomUnit.retrieve(
            self.entity,
            self.session,
            self.custom_unit_id
        )

    @classmethod
    def _decode_entries(cls: Type[T], data: Any) -> List[Entry]:
        """Return Entries decoded from API response data"""
        if not isinstance(data, list):
            raise ApiError('Unexpected API response type ' + str(type(data)))

        def decode(obj) -> Entry:
            if not isinstance(obj, dict):
                raise ApiError('Unexpected API object type ' + str(type(obj)))

            try:
                entry = Entry(
                    side=Side(obj['side']),
                    amount=AmatinoAmount.decode(obj['amount']),
                    account_id=obj['account_id'],
                    description=obj['description']
                )
            except KeyError as error:
                raise MissingKey(error.args[0])

            return entry

        return [decode(e) for e in data]

    class RetrieveArguments(Encodable):
        def __init__(
            self,
            transaction_id: int,
            denomination: Denomination,
            version: Optional[int]
        ) -> None:

            if not isinstance(transaction_id, int):
                raise TypeError('transaction_id must be of type `int`')

            if not isinstance(denomination, Denomination):
                raise TypeError('denomination must be of type `Denomination`')

            if version and not isinstance(version, int):
                raise TypeError('version must be of type `Optional[int]`')

            self._transaction_id = transaction_id

            if isinstance(denomination, GlobalUnit):
                self._global_unit_id = denomination.id_
                self._custom_unit_id = None
            else:
                self._global_unit_id = None
                self._custom_unit_id = denomination.id_

            self._version = version

            return

        def serialise(self) -> Dict[str, Any]:
            data = {
                'transaction_id': self._transaction_id,
                'global_unit_denomination': self._global_unit_id,
                'custom_unit_denomination': self._custom_unit_id,
                'version': self._version
            }
            return data

    class CreateArguments(Encodable):
        def __init__(
            self,
            time: datetime,
            entries: List[Entry],
            denomination: Denomination,
            description: Optional[str] = None,
        ) -> None:

            if not isinstance(time, datetime):
                raise TypeError('time must be of type `datetime.datetime`')

            if not isinstance(denomination, Denomination):
                raise TypeError('denomination must be of type `Denomination`')

            self._time = AmatinoTime(time)
            self._denomination = denomination
            self._entries = Transaction._Entries(entries)
            self._description = Transaction._Description(description)

            return

        def serialise(self) -> Dict[str, Any]:
            if isinstance(self._denomination, CustomUnit):
                custom_unit_id = self._denomination.id_
                global_unit_id = None
            else:
                custom_unit_id = None
                global_unit_id = self._denomination.id_

            data = {
                'transaction_time': self._time.serialise(),
                'custom_unit_denomination': custom_unit_id,
                'global_unit_denomination': global_unit_id,
                'description': self._description.serialise(),
                'entries': self._entries.serialise()
            }
            return data

    class UpdateArguments(Encodable):
        def __init__(
            self,
            transaction: T,
            time: Optional[datetime] = None,
            entries: Optional[List[Entry]] = None,
            denomination: Optional[Denomination] = None,
            description: Optional[str] = None
        ) -> None:

            if not isinstance(transaction, Transaction):
                raise TypeError('transaction must be of type `Transaction`')

            self._transaction_id = transaction.id_

            if time:
                if not isinstance(time, datetime):
                    raise TypeError('time must be of type `datetime`')
                self._ = AmatinoTime(time)
            else:
                self._time = AmatinoTime(transaction.time)

            if entries:
                self._entries = Transaction._Entries(entries)
            else:
                self._entries = Transaction._Entries(transaction.entries)

            if denomination:
                if not isinstance(denomination, Denomination):
                    raise TypeError(
                        'demomination must be of type `Denomination`'
                    )
                self._denomination = denomination
            else:
                self._denomination = transaction.denomination

            if description:
                self._description = Transaction._Description(description)
            else:
                self._description = Transaction._Description(
                    transaction.description
                )

            return

        def serialise(self) -> Dict[str, Any]:
            if isinstance(self._denomination, CustomUnit):
                custom_unit_id = self._denomination.id_
                global_unit_id = None
            else:
                custom_unit_id = None
                global_unit_id = self._denomination.id_

            data = {
                'transaction_id': self._transaction_id,
                'transaction_time': self._time.serialise(),
                'custom_unit_denomination': custom_unit_id,
                'global_unit_denomination': global_unit_id,
                'description': self._description.serialise(),
                'entries': self._entries.serialise()
            }
            return data

    class _Description(ConstrainedString):
        def __init__(self, string: Optional[str] = None) -> None:
            if string is None:
                string = ''
            super().__init__(
                string,
                'description',
                Transaction.MAX_DESCRIPTION_LENGTH
            )
            return

    class _Entries(Encodable):
        def __init__(self, entries: List[Entry]) -> None:

            if not isinstance(entries, list):
                raise TypeError('entries must be of type List[Entry]')
            if False in [isinstance(e, Entry) for e in entries]:
                raise TypeError('entries must be of type List[Entry]')

            debits = sum([e.amount for e in entries if e.side == Side.debit])
            credits_ = sum([e.amount for e in entries if e.side == Side.credit])

            if debits != credits_:
                raise ValueError('sum of debits must equal sum of credits')

            self._entries = entries
            return

        entries = Immutable(lambda s: s._entries)
        debits = Immutable(
            lambda s: [e for e in s._entries if e.side == Side.debit]
        )
        credits_ = Immutable(
            lambda s: [e for e in s._entries if e.side == Side.credit]
        )

        def serialise(self) -> List[Dict]:
            return [e.serialise() for e in self._entries]



"""
Amatino API Python Bindings
Performance Module
Author: hugh@amatino.io
"""
from typing import Type
from typing import TypeVar
from typing import List
from typing import Optional
from typing import Any
from typing import Dict
from datetime import datetime
from amatino.denominated import Denominated
from amatino.denomination import Denomination
from amatino.decodable import Decodable
from amatino.internal.data_package import DataPackage
from amatino.internal.api_request import ApiRequest
from amatino.internal.url_parameters import UrlParameters
from amatino.internal.http_method import HTTPMethod
from amatino.internal.encodable import Encodable
from amatino.missing_key import MissingKey
from amatino.entity import Entity
from amatino.unexpected_response_type import UnexpectedResponseType
from amatino.internal.am_time import AmatinoTime
from amatino.tree_node import TreeNode
from amatino.internal.immutable import Immutable
from amatino.global_unit import GlobalUnit
from amatino.custom_unit import CustomUnit

T = TypeVar('T', bound='Performance')
K = TypeVar('K', bound='Performance.RetrieveArguments')


class Performance(Denominated, Decodable):
    """
    A Performance is a hierarchical collection of Account balances describing
    the financial performance of an Entity over a period of time. They are
    generic representations of popular accounting constructs known as the
    'Income Statement', 'Profit & Loss', or 'Statement of Financial
    Performance'.

    The Peformance object is jurisdiction agnostic, and obeys simple
    double-entry accounting rules. They list income and expense, each nesting
    its own children.

    You can retrieve a Performance denominated in an arbitrary Global Unit or
    Custom Unit. Amatino will automatically calculate the implicit gain or loss
    relative to each Account's underlying denomination and include those gains
    and losses in each Account balance.

    A Performance may be retrieved to an arbitrary depth. Depth in the
    Performance context is the number of levels down the Account hierarchy
    Amatino should go when retrieving the Performance. For example, if a
    top-level Account has child accounts three layers deep, then specifying a
    depth of three will retrieve all those children.

    Regardless of the depth you specify, Amatino will calculate recursive
    balances at full depth.
    """

    _PATH = '/performances'

    def __init__(
        self,
        entity: Entity,
        start_time: AmatinoTime,
        end_time: AmatinoTime,
        generated_time: AmatinoTime,
        global_unit_id: Optional[int],
        custom_unit_id: Optional[int],
        income: Optional[List[TreeNode]],
        expenses: Optional[List[TreeNode]],
        depth: int
    ) -> None:

        assert isinstance(entity, Entity)
        assert isinstance(start_time, AmatinoTime)
        assert isinstance(end_time, AmatinoTime)
        assert isinstance(generated_time, AmatinoTime)
        if global_unit_id is not None:
            assert isinstance(global_unit_id, int)
        if custom_unit_id is not None:
            assert isinstance(custom_unit_id, int)
        if income is not None:
            assert isinstance(income, list)
            assert False not in [isinstance(i, TreeNode) for i in income]
        if expenses is not None:
            assert isinstance(expenses, list)
            assert False not in [isinstance(e, TreeNode) for e in expenses]
        assert isinstance(depth, int)

        self._entity = entity
        self._start_time = start_time
        self._end_time = end_time
        self._generated_time = generated_time
        self._global_unit_id = global_unit_id
        self._custom_unit_id = custom_unit_id
        self._income = income
        self._expenses = expenses
        self._depth = depth

        return

    entity = Immutable(lambda s: s._entity)
    session = Immutable(lambda s: s._entity.session)
    start_time = Immutable(lambda s: s._start_time.raw)
    end_time = Immutable(lambda s: s._end_time.raw)
    generated_time = Immutable(lambda s: s._generated_time.raw)
    custom_unit_id = Immutable(lambda s: s._custom_unit_id)
    global_unit_id = Immutable(lambda s: s._global_unit_id)
    income = Immutable(lambda s: s._income)
    expenses = Immutable(lambda s: s._expenses)

    has_income = Immutable(
        lambda s: s._income is not None and len(s._income) > 0
    )
    has_expenses = Immutable(
        lambda s: s._expenses is not None and len(s._expenses) > 0
    )

    @classmethod
    def decode(
        cls: Type[T],
        entity: Entity,
        data: Any
    ) -> T:

        if not isinstance(data, dict):
            raise UnexpectedResponseType(data, dict)

        try:

            income = None
            if data['income'] is not None:
                income = TreeNode.decode_many(entity, data['income'])

            expenses = None
            if data['expenses'] is not None:
                expenses = TreeNode.decode_many(entity, data['expenses'])

            performance = cls(
                entity=entity,
                start_time=AmatinoTime.decode(data['start_time']),
                end_time=AmatinoTime.decode(data['end_time']),
                generated_time=AmatinoTime.decode(data['generated_time']),
                custom_unit_id=data['custom_unit_denomination'],
                global_unit_id=data['global_unit_denomination'],
                income=income,
                expenses=expenses,
                depth=data['depth']
            )
        except KeyError as error:
            raise MissingKey(error.args[0])

        return performance

    @classmethod
    def retrieve(
        cls: Type[T],
        entity: Entity,
        start_time: datetime,
        end_time: datetime,
        denomination: Denomination,
        depth: Optional[int] = None
    ) -> T:

        arguments = cls.RetrieveArguments(
            start_time=start_time,
            end_time=end_time,
            denomination=denomination,
            depth=depth
        )

        return cls._retrieve(entity, arguments)

    @classmethod
    def _retrieve(cls: Type[T], entity: Entity, arguments: K) -> T:
        """Retrieve a Performance"""
        if not isinstance(entity, Entity):
            raise TypeError('entity must be of type `Entity`')

        if not isinstance(arguments, cls.RetrieveArguments):
            raise TypeError(
                'arguments must be of type Performance.RetrieveArguments'
            )

        data = DataPackage(object_data=arguments, override_listing=True)
        parameters = UrlParameters(entity_id=entity.id_)

        request = ApiRequest(
            path=cls._PATH,
            method=HTTPMethod.GET,
            credentials=entity.session,
            data=data,
            url_parameters=parameters
        )

        return cls.decode(entity, request.response_data)

    class RetrieveArguments(Encodable):
        def __init__(
            self,
            start_time: datetime,
            end_time: datetime,
            denomination: Denomination,
            depth: Optional[int] = None
        ) -> None:

            if not isinstance(start_time, datetime):
                raise TypeError('start_time must be of type `datetime`')

            if not isinstance(end_time, datetime):
                raise TypeError('end_time must be of type `datetime`')

            if not isinstance(denomination, Denomination):
                raise TypeError('denomination must be of type `Denomination`')

            if depth is not None and not isinstance(depth, int):
                raise TypeError(
                    'If supplied, depth must be of type `int`'
                )

            self._start_time = AmatinoTime(start_time)
            self._end_time = AmatinoTime(end_time)
            self._denomination = denomination
            self._depth = depth

            return

        def serialise(self) -> Dict[str, Any]:

            global_unit_id = None
            custom_unit_id = None

            if isinstance(self._denomination, GlobalUnit):
                global_unit_id = self._denomination.id_
            else:
                assert isinstance(self._denomination, CustomUnit)
                custom_unit_id = self._denomination.id_

            data = {
                'start_time': self._start_time.serialise(),
                'end_time': self._end_time.serialise(),
                'custom_unit_denomination': custom_unit_id,
                'global_unit_denomination': global_unit_id,
                'depth': self._depth
            }

            return data



"""
Amatino API Python Bindings
Amatino Error Module
Author: hugh@amatino.io
"""


class AmatinoError(Exception):
    """
    Base exception class for all errors thrown by the Amatino library
    """
    pass



"""
Amatino API Python Bindings
Constrained Integer Module
Author: hugh@amatino.io

This module is intended to be private, used indirectly by public classes, and
should not be used directly.

"""
from typing import Optional
from amatino.constraint_error import ConstraintError
from amatino.internal.encodable import Encodable


class ConstrainedInteger(Encodable):
    """
    An integer whose maximum, and optionally minimum, magnitude is restricted.
    """
    MAX_ERR = '{name} above maximum value of {}'
    MIN_ERR = '{name} below minimum value of {min_value}'

    def __init__(
        self,
        integer: int,
        name: str,
        max_value: int,
        min_value: Optional[int] = None
    ) -> None:

        assert isinstance(name, str)
        assert isinstance(max_value, int)
        if min_value is not None:
            assert isinstance(min_value, int)

        if not isinstance(integer, int):
            raise TypeError(name + ' must be of type `int`')

        if integer > max_value:
            error = self.MAX_ERR.format(name=name, max_value=str(max_value))
            raise ConstraintError(error)

        self._integer = integer

        if min_value is None:
            return

        if integer < min_value:
            error = self.MIN_ERR.format(name=name, min_value=str(min_value))
            raise ConstraintError(error)

        return

    def serialise(self) -> int:
        return self._integer



"""
Amatino API Python Bindings
Account Test Module
Author: hugh@amatino.io
"""
from amatino import Account
from amatino import Entity
from amatino import AMType
from amatino import GlobalUnit
from amatino.tests.primary.entity import EntityTest
from amatino.internal.immutable import Immutable

USD_UNIT_ID = 5


class AccountTest(EntityTest):
    """Test the Account primary object"""

    def __init__(self, name='Create, retrieve, update an Account') -> None:

        super().__init__(name)
        self.create_entity()
        if not isinstance(self.entity, Entity):
            raise RuntimeError(
                'Entity creation failed. Consider running Entity tests'
            )
        return

    usd = Immutable(lambda s: s._usd())
    _account_test_cached_usd = None

    def _usd(self) -> GlobalUnit:
        if self._account_test_cached_usd is not None:
            return self._account_test_cached_usd
        usd = GlobalUnit.retrieve(self.session, USD_UNIT_ID)
        self._account_test_cached_usd = usd
        return usd

    def create_account(self, amt=AMType.asset, name='Test account') -> Account:

        account = Account.create(
            self.session,
            self.entity,
            name,
            amt,
            self.usd,
            'A test Account created by the Python test suite'
        )

        self.account = account

        return account

    def execute(self) -> None:

        try:
            account = self.create_account()
        except Exception as error:
            self.record_failure(error)
            return

        assert isinstance(account, Account)

        try:
            account = Account.retrieve(
                self.session,
                self.entity,
                account.id_
            )
        except Exception as error:
            self.record_failure(error)
            return

        if account.id_ != self.account.id_:
            self.record_failure('Account IDs do not match')
            return

        new_name = 'Updated account name'

        try:
            updated_account = account.update(name=new_name)
        except Exception as error:
            self.record_failure(error)
            return

        if updated_account.name != new_name:
            self.record_failure('Account name was not updated')
            return

        self.record_success()
        return



"""
Amatino API Python Bindings
Amatino Amount Module
Author: hugh@amatino.io
"""

from decimal import Decimal
from typing import TypeVar
from typing import Type

T = TypeVar('T', bound='AmatinoAmount')


class AmatinoAmount(Decimal):
    """
    Internal class bridging string amount representations to the Python
    Decimal class.
    """

    @classmethod
    def decode(cls: Type[T], amount: str) -> T:
        """Return a Decimal number decoded from an API response"""
        assert isinstance(amount, str)

        negate = False
        if amount[0] == "(":
            amount = amount[1:-1]
            negate = True

        value = cls(amount.replace(",", ""))

        if negate is True:
            value = value * -1

        return value



"""
Amatino API Python Bindings
Ledger Row Module
Author: hugh@amatino.io
"""
from amatino.session import Session
from amatino.entity import Entity
from amatino.account import Account
from amatino.internal.am_time import AmatinoTime
from amatino.transaction import Transaction
from amatino.internal.immutable import Immutable
from decimal import Decimal
from typing import Optional


class LedgerRow:
    """
    A Ledger Row is a specialised view of a Transaction, delivered as part of a
    Ledger or Recursive Ledger. The Ledger Row describes a Tranasction from the
    perspective of the Account targeted by the controlling Ledger or Recursive
    Ledger.

    When consuming the Amatino API, you will never encounter a Ledger Row on its
    own. They are only ever delivered under the ledger_rows key as part of a
    Ledger or Recursive Ledger object.
    """

    def __init__(
        self,
        transaction_id: int,
        transaction_time: AmatinoTime,
        description: str,
        opposing_account_id: Optional[int],
        opposing_account_name: str,
        debit: Decimal,
        credit: Decimal,
        balance: Decimal
    ) -> None:

        self._transaction_id = transaction_id
        self._transaction_time = transaction_time
        self._description = description
        self._opposing_account_id = opposing_account_id
        self._opposing_account_name = opposing_account_name
        self._debit = debit
        self._credit = credit
        self._balance = balance

        return

    transaction_id = Immutable(lambda s: s._transaction_id)
    transaction_time = Immutable(lambda s: s._transaction_time.raw)
    description = Immutable(lambda s: s._description)
    opposing_account_id = Immutable(lambda s: s._opposing_account_id)
    opposing_account_name = Immutable(lambda s: s._opposing_account_name)
    debit = Immutable(lambda s: s._debit)
    credit = Immutable(lambda s: s._credit)
    balance = Immutable(lambda s: s._balance)

    def opposing_account(
        self,
        session: Session,
        entity: Entity
    ) -> Optional[Account]:
        """Return the account opposing this transaction, or, if many, None"""
        if self.opposing_account_id is None:
            return None
        return Account.retrieve(session, entity, self.opposing_account_id)

    def retrieve_transaction(
        self,
        session: Session,
        entity: Entity
    ) -> Transaction:
        """Retrieve the Transaction this LedgerRow describes"""
        return Transaction.retrieve(session, entity, self.transaction_id)



"""
Amatino API
Tree Module
Author: hugh@amatino.io
"""
from typing import Type
from typing import TypeVar
from typing import List
from typing import Optional
from typing import Any
from typing import Dict
from datetime import datetime
from amatino.denominated import Denominated
from amatino.denomination import Denomination
from amatino.decodable import Decodable
from amatino.internal.data_package import DataPackage
from amatino.internal.api_request import ApiRequest
from amatino.internal.url_parameters import UrlParameters
from amatino.internal.http_method import HTTPMethod
from amatino.internal.encodable import Encodable
from amatino.missing_key import MissingKey
from amatino.entity import Entity
from amatino.unexpected_response_type import UnexpectedResponseType
from amatino.internal.am_time import AmatinoTime
from amatino.tree_node import TreeNode
from amatino.internal.immutable import Immutable
from amatino.global_unit import GlobalUnit
from amatino.custom_unit import CustomUnit

T = TypeVar('T', bound='Tree')


class Tree(Decodable, Denominated):
    """
    Trees present the entire chart of Accounts of an Entity in a single
    hierarchical object.

    Each Account is nested under its parent, and in turn lists all its
    children, all providing Balances and Recursive Balances.

    Trees are trimmed for permissions. If the user requesting the tree only
    has read access to a subset of Accounts, they will only receive a tree
    containing those accounts, with placeholder objects filling the place
    of Accounts they are not permitted to read.

    Each Account in the Tree is presented as a Tree Node.
    """
    _PATH = '/trees'

    def __init__(
        self,
        entity: Entity,
        balance_time: AmatinoTime,
        generated_time: AmatinoTime,
        global_unit_denomination: Optional[int],
        custom_unit_denomination: Optional[int],
        tree: Optional[List[TreeNode]]
    ) -> None:

        assert isinstance(entity, Entity)
        assert isinstance(balance_time, AmatinoTime)
        assert isinstance(generated_time, AmatinoTime)
        if global_unit_denomination is not None:
            assert isinstance(global_unit_denomination, int)
        if custom_unit_denomination is not None:
            assert isinstance(custom_unit_denomination, int)
        if tree is not None:
            assert isinstance(tree, list)
            assert False not in [isinstance(t, TreeNode) for t in tree]
        
        self._entity = entity
        self._balance_time = balance_time
        self._generated_time = generated_time
        self._global_unit_id = global_unit_denomination
        self._custom_unit_id = custom_unit_denomination
        self._tree = tree

        return

    entity = Immutable(lambda s: s._entity)
    session = Immutable(lambda s: s._entity.session)
    start_time = Immutable(lambda s: s._start_time)
    end_time = Immutable(lambda s: s._end_time)
    generated_time = Immutable(lambda s: s._generated_time)
    custom_unit_id = Immutable(lambda s: s._custom_unit_id)
    global_unit_id = Immutable(lambda s: s._global_unit_id)
    nodes = Immutable(lambda s: s._tree)

    has_accounts = Immutable(
        lambda s: s._nodes is not None and len(s._nodes) > 0
    )

    @classmethod
    def decode(cls: Type[T], entity: Entity, data: Any) -> T:

        if not isinstance(data, dict):
            raise UnexpectedResponseType(data, dict)

        try:
            tree = None
            if data['tree']


"""
Amatino API Python Bindings
Balance Module
Author: hugh@amatino.io
"""
from amatino.internal.immutable import Immutable
from amatino.internal.balance_protocol import BalanceProtocol


class Balance(BalanceProtocol):
    """
    A Balance represents the sum total value of all Entries party to an Account.
    """

    PATH = '/accounts/balance'



"""
Amatino API Python Bindings
Test Sequence Module
Author: hugh@amatino.io

Provides a constant manifest of tests to execute
"""
import amatino.tests.primary as primary
import amatino.tests.ancillary as ancillary
import amatino.tests.derived as derived
from amatino.tests.ancillary.custom_unit import CustomUnitTest

SEQUENCE = [
    ancillary.SessionTest,
    ancillary.GlobalUnitTest,
    CustomUnitTest,
    ancillary.UserTest,
    primary.EntityTest,
    primary.AccountTest,
    primary.TransactionTest,
    derived.LedgerTest,
    derived.RecursiveLedgerTest,
    derived.BalanceTest,
    derived.RecursiveBalanceTest,
    derived.PositionTest,
    derived.PerformanceTest
]



"""
Procuret API Python Bindings
Position Module
Author: hugh@blinkybeach.com
"""
from typing import Type
from typing import TypeVar
from typing import List
from typing import Optional
from typing import Any
from typing import Dict
from datetime import datetime
from amatino.denominated import Denominated
from amatino.denomination import Denomination
from amatino.decodable import Decodable
from amatino.internal.data_package import DataPackage
from amatino.internal.api_request import ApiRequest
from amatino.internal.url_parameters import UrlParameters
from amatino.internal.http_method import HTTPMethod
from amatino.internal.encodable import Encodable
from amatino.missing_key import MissingKey
from amatino.entity import Entity
from amatino.unexpected_response_type import UnexpectedResponseType
from amatino.internal.am_time import AmatinoTime
from amatino.tree_node import TreeNode
from amatino.internal.immutable import Immutable
from amatino.global_unit import GlobalUnit
from amatino.custom_unit import CustomUnit

T = TypeVar('T', bound='Position')
K = TypeVar('K', bound='Position.RetrieveArguments')


class Position(Denominated, Decodable):
    """
    Positions are hierarchical collections of Account balances describing
    the financial position of an Entity at a point in time. They are generic representations of popular accounting constructs better known as a
    'Balance Sheet', 'Statement of Financial Position', or 'Statements of
    Assets, Liabilities and Owner's Equity.

    Positions are jurisdiction agnostic, and obey simple double-entry
    accounting rules. They list asset, liability, and equity Accounts, each
    nesting its own children.,

    You can retrieve Positions denominated in arbitrary Global Unit or
    Custom Unit. When a particular denomination gives rise to an unrealised
    gain or loss, Amatino will automatically calculate that gain or loss
    and include it as a top-level account in the returned Position.

    Positions may also be retrieved to an arbitrary depth. Depth in the
    Position context is the number of levels down the Account hierarchy
    Amatino should go when retrieving the Position. For example, if a
    top-level Account has child accounts three layers deep, then specifying
    a depth of three will retrieve all those children.

    Regardless of the depth you specify, Amatino will calculate recursive
    balances for all Accounts at full depth.
    """

    _PATH = '/positions'

    def __init__(
        self,
        entity: Entity,
        balance_time: AmatinoTime,
        generated_time: AmatinoTime,
        global_unit_id: Optional[int],
        custom_unit_id: Optional[int],
        assets: Optional[List[TreeNode]],
        liabilities: Optional[List[TreeNode]],
        equities: Optional[List[TreeNode]],
        depth: int
    ) -> None:

        assert isinstance(entity, Entity)
        assert isinstance(balance_time, AmatinoTime)
        assert isinstance(generated_time, AmatinoTime)
        if global_unit_id is not None:
            assert isinstance(global_unit_id, int)
        if custom_unit_id is not None:
            assert isinstance(custom_unit_id, int)
        if assets is not None:
            assert isinstance(assets, list)
            assert False not in [isinstance(a, TreeNode) for a in assets]
        if liabilities is not None:
            assert isinstance(liabilities, list)
            assert False not in [isinstance(l, TreeNode) for l in liabilities]
        if equities is not None:
            assert isinstance(equities, list)
            assert False not in [isinstance(e, TreeNode) for e in equities]

        self._entity = entity
        self._balance_time = balance_time
        self._generated_time = generated_time
        self._global_unit_id = global_unit_id
        self._custom_unit_id = custom_unit_id
        self._assets = assets
        self._liabilities = liabilities
        self._equities = equities
        self._depth = depth

        return

    session = Immutable(lambda s: s._entity.session)
    entity = Immutable(lambda s: s._entity)
    balance_time = Immutable(lambda s: s._balance_time.raw)
    generated_time = Immutable(lambda s: s._generated_time.raw)
    custom_unit_id = Immutable(lambda s: s._custom_unit_id)
    global_unit_id = Immutable(lambda s: s._global_unit_id)
    assets = Immutable(lambda s: s._assets)
    liabilities = Immutable(lambda s: s._liabilities)
    equities = Immutable(lambda s: s._equities)

    has_assets = Immutable(
        lambda s: s._assets is not None and len(s._assets) > 0
    )
    has_liabilities = Immutable(
        lambda s: s._liabilities is not None and len(s._liabilities) > 0
    )
    has_equities = Immutable(
        lambda s: s._equities is not None and len(s._equities) > 0
    )

    @classmethod
    def decode(
        cls: Type[T],
        entity: Entity,
        data: Any
    ) -> T:

        if not isinstance(data, dict):
            raise UnexpectedResponseType(data, dict)

        try:

            assets = None
            if data['assets'] is not None:
                assets = TreeNode.decode_many(entity, data['assets'])

            liabilities = None
            if data['liabilities'] is not None:
                liabilities = TreeNode.decode_many(
                    entity,
                    data['liabilities']
                )

            equities = None
            if data['equities'] is not None:
                equities = TreeNode.decode_many(entity, data['equities'])

            position = cls(
                entity=entity,
                balance_time=AmatinoTime.decode(data['balance_time']),
                generated_time=AmatinoTime.decode(data['generated_time']),
                global_unit_id=data['global_unit_denomination'],
                custom_unit_id=data['custom_unit_denomination'],
                assets=assets,
                liabilities=liabilities,
                equities=equities,
                depth=data['depth']
            )

        except KeyError as error:
            raise MissingKey(error.args[0])

        return position

    @classmethod
    def retrieve(
        cls: Type[T],
        entity: Entity,
        balance_time: datetime,
        denomination: Denomination,
        depth: Optional[int] = None
    ) -> T:

        arguments = cls.RetrieveArguments(
            balance_time=balance_time,
            denomination=denomination,
            depth=depth
        )

        return cls._retrieve(entity, arguments)

    @classmethod
    def _retrieve(cls: Type[T], entity: Entity, arguments: K) -> T:
        """Retrieve a Position"""
        if not isinstance(entity, Entity):
            raise TypeError('entity must be of type `Entity`')

        if not isinstance(arguments, cls.RetrieveArguments):
            raise TypeError(
                'arguments must be of type Position.RetrieveArguments'
            )

        data = DataPackage(object_data=arguments, override_listing=True)
        parameters = UrlParameters(entity_id=entity.id_)

        request = ApiRequest(
            path=cls._PATH,
            method=HTTPMethod.GET,
            credentials=entity.session,
            data=data,
            url_parameters=parameters
        )

        return cls.decode(entity, request.response_data)

    class RetrieveArguments(Encodable):
        def __init__(
            self,
            balance_time: datetime,
            denomination: Denomination,
            depth: Optional[int] = None
        ) -> None:

            if not isinstance(balance_time, datetime):
                raise TypeError('balance_time must be of type `datetime`')

            if not isinstance(denomination, Denomination):
                raise TypeError('denomination must be of type `Denomination`')

            if depth is not None and not isinstance(depth, int):
                raise TypeError('If supplied, depth must be `int`')

            self._balance_time = AmatinoTime(balance_time)
            self._denomination = denomination
            self._depth = depth

            return

        def serialise(self) -> Dict[str, Any]:

            global_unit_id = None
            custom_unit_id = None

            if isinstance(self._denomination, GlobalUnit):
                global_unit_id = self._denomination.id_
            else:
                assert isinstance(self._denomination, CustomUnit)
                custom_unit_id = self._denomination.id_

            data = {
                'balance_time': self._balance_time.serialise(),
                'custom_unit_denomination': custom_unit_id,
                'global_unit_denomination': global_unit_id,
                'depth': self._depth
            }

            return data



"""
Amatino API Python Bindings
Region Module
Author: hugh@amatino.io
"""
from amatino.internal.immutable import Immutable


class Region:
    """
    A geographic region in which Amatino can store accounting information.
    """
    def __init__(
        self,
        id_: int,
        region_code: str
    ) -> None:
        self._id = id_
        self._region_code = region_code
        return

    id_ = Immutable(lambda s: s._id)



"""
Amatino API Python Bindings
Color Module
Author: hugh@amatino.io
"""
from typing import TypeVar
from typing import Type
from amatino.internal.immutable import Immutable
from amatino.internal.encodable import Encodable

T = TypeVar('T', bound='Color')


class Color:
    """
    A representation of an RGB color.
    """
    def __init__(
        self,
        red: int,
        green: int,
        blue: int
    ) -> None:

        if (
                not isinstance(red, int)
                or not isinstance(green, int)
                or not isinstance(blue, int)
        ):
            raise TypeError('Color components must be of type int')

        if (
                red < 0 or red > 255 or
                green < 0 or green > 255 or
                blue < 0 or blue > 255
        ):
            raise ValueError('Color components must be >= 0 and <= 255')

        self._red = red
        self._green = green
        self._blue = blue

        return

    red = Immutable(lambda s: s._red)
    green = Immutable(lambda s: s._green)
    blue = Immutable(lambda s: s._blue)
    hex_string = Immutable(lambda s: s.as_hex_string())

    @classmethod
    def from_hex_string(cls: Type[T], hexstring: str) -> T:

        if not isinstance(hexstring, str):
            raise TypeError('hexstring must be of type `str`')

        if len(hexstring) != 6:
            raise ValueError('hexstring must be 6 characters')

        red = int(hexstring[:2], 16)
        green = int(hexstring[2:4], 16)
        blue = int(hexstring[4:6], 16)

        color = cls(red=red, green=green, blue=blue)

        return color

    def as_hex_string(self) -> str:
        """
        Return the colour as a hex value string
        """
        hex_string = hex(self.red)[2:]
        hex_string += hex(self.green)[2:]
        hex_string += hex(self.blue)[2:]
        return hex_string

    def as_int_tuple(self) -> tuple:
        """
        Return the colour as a tuple of three integers, in the order
        (red, green, blue)
        """
        return (self._red, self._green, self._blue)

    def serialise(self) -> str:
        assert len(self.hex_string) == 6
        return self.hex_string



"""
Amatino API Python Bindings
AM Side Module
Author: hugh@amatino.io
"""
from enum import Enum


class Side(Enum):
    """
    Side is Amatino's name for the fundamental double-entry bookeeping
    building blocks, the debit and credit. They are constant and unchanging, so
    while you can retrieve them with a GET request, you can also safely
    hard-code them into your application.

    When consuming the Amatino API, Sides primarily pop up when dealing with
    Transactions. Transactions are made up of multiple Entries, each of which
    must include a Side.
    """
    debit = 0
    credit = 1



"""
Amatino API Python Bindings
Entity Test Module
Author: hugh@amatino.io
"""
from amatino.entity import Entity
from amatino.tests.ancillary.session import SessionTest
from amatino import Session


class EntityTest(SessionTest):
    """
    Test the Entity primary object
    """

    def __init__(self, name='Create, retrieve, update an Entity') -> None:

        self.entity = None

        super().__init__(name)
        self.create_session()
        if not isinstance(self.session, Session):
            raise RuntimeError(
                'Session creation failed, consider running Session tests'
            )
        return

    def create_entity(self) -> Entity:
        entity = Entity.create(
            self.session,
            'Test Entity',
            None,
            None
        )
        self.entity = entity
        return entity

    def execute(self) -> None:

        assert self.session is not None

        try:
            entity = self.create_entity()
        except Exception as error:
            self.record_failure(error)
            return

        assert isinstance(self.entity, Entity)

        try:
            entity = Entity.retrieve(
                self.session,
                entity.id_
            )
        except Exception as error:
            self.record_failure(error)
            return

        if entity.id_ != self.entity.id_:
            self.record_failure('Entity ids do not match')
            return

        new_name = 'Updated Entity Name'

        try:
            updated_entity = entity.update(new_name)
        except Exception as error:
            self.record_failure(error)
            return

        if updated_entity.name != new_name:
            self.record_failure('Entity name not updated: ' + str(entity.name))
            return

        self.record_success()
        return



"""
Amatino API Python Bindings
Tree Node Module
Author: hugh@amatino.io
"""
from typing import TypeVar
from typing import Type
from typing import Any
from typing import List
from typing import Optional
from amatino.decodable import Decodable
from amatino.missing_key import MissingKey
from amatino.am_type import AMType
from amatino.entity import Entity
from amatino.internal.immutable import Immutable
from amatino.unexpected_response_type import UnexpectedResponseType
from amatino.account import Account
from amatino.internal.am_amount import AmatinoAmount
from decimal import Decimal

T = TypeVar('T', bound='TreeNode')


class TreeNode(Decodable):
    """
    A Tree Node is a specialised view of an Account. It provides a recursive
    and individual balance for an Account. You will never interact with Tree
    Nodes' directly, instead you will receive lists of Tree Nodes as components
    of requests for Trees, Positions, and Performances.

    Tree Nodes are recursive objects that may contain other Tree Nodes,
    depending on whether the Account in question has any children.

    Trees, Positions, and Performances describe entire Entities. The User from
    whose perspective you retrieve a Tree, Position, or Performance may not
    have read access to every Account in the Entity. In such cases, Tree Nodes
    describing Accounts to which a User does not have read access will be
    returned with null in their balance fields, and a generic Type in place of
    the actual Account name.
    """

    def __init__(
        self,
        entity: Entity,
        account_id: int,
        depth: int,
        account_balance: Decimal,
        recursive_balance: Decimal,
        name: str,
        am_type: AMType,
        children: Optional[List[T]]
    ) -> None:

        assert isinstance(entity, Entity)
        assert isinstance(account_id, int)
        assert isinstance(depth, int)
        assert isinstance(account_balance, Decimal)
        assert isinstance(recursive_balance, Decimal)
        assert isinstance(name, str)
        assert isinstance(am_type, AMType)
        if children is not None:
            assert isinstance(children, list)
            assert False not in [isinstance(c, TreeNode) for c in children]

        self._entity = entity
        self._account_id = account_id
        self._depth = depth
        self._account_balance = account_balance
        self._recursive_balance = recursive_balance
        self._name = name
        self._am_type = am_type
        self._children = children

        return

    session = Immutable(lambda s: s._entity.session)
    entity = Immutable(lambda s: s._entity)
    account_id = Immutable(lambda s: s._account_id)
    depth = Immutable(lambda s: s._depth)
    account_balance = Immutable(lambda s: s._account_balance)
    recursive_balance = Immutable(lambda s: s._recursive_balance)
    name = Immutable(lambda s: s._name)
    am_type = Immutable(lambda s: s._am_type)
    children = Immutable(lambda s: s._children)

    has_children = Immutable(
        lambda s: s._children is not None and len(s._children) > 0
    )
    account = Immutable(lambda s: s._account())

    _node_cached_account = None

    def _account(self) -> Account:
        """
        Return the Account this TreeNode describes. Cache it for repeated
        requests.
        """
        if isinstance(self._node_cached_account, Account):
            return self._node_cached_account
        account = Account.retrieve(
            session=self.entity.session,
            entity=self.entity,
            account_id=self.account_id
        )
        self._node_cached_account = account
        return account

    @classmethod
    def decode(cls: Type[T], entity: Entity, data: Any) -> T:

        if not isinstance(data, dict):
            raise UnexpectedResponseType(data, dict)

        try:
            children = None
            if data['children'] is not None:
                children = [cls.decode(entity, d) for d in data['children']]
            node = cls(
                entity=entity,
                account_id=data['account_id'],
                depth=data['depth'],
                account_balance=AmatinoAmount.decode(
                    data['account_balance']
                ),
                recursive_balance=AmatinoAmount.decode(
                    data['recursive_balance']
                ),
                name=data['name'],
                am_type=AMType(data['type']),
                children=children
            )
        except KeyError as error:
            raise MissingKey(error.args[0])

        return node



"""
Amatino API Python Bindings
Custom Unit Test Module
Author: hugh@amatino.io
"""
from amatino.tests.primary.entity import EntityTest
from amatino import CustomUnit


class CustomUnitTest(EntityTest):
    """Test the Custom Unit object"""

    def __init__(self, name='Create, retrieve, update a Custom Unit') -> None:

        super().__init__(name)
        self.create_entity()
        return

    def execute(self) -> None:

        try:
            custom_unit = CustomUnit.create(
                self.entity,
                self.session,
                'Bojangles',
                'BTX',
                4
            )
        except Exception as error:
            self.record_failure(error)
            return

        if not isinstance(custom_unit, CustomUnit):
            return_type = str(type(custom_unit))
            self.record_failure('Unexpected return type: ' + return_type)
            return

        if not isinstance(custom_unit.id_, int):
            id_type = str(type(custom_unit.id_))
            self.record_failure('Unexpected id type: ' + id_type)
            return

        if not isinstance(custom_unit.code, str):
            code_type = str(type(custom_unit.code))
            self.record_failure('Unexpected code type: ' + code_type)
            return

        try:
            retrieved_custom_unit = CustomUnit.retrieve(
                self.entity,
                self.session,
                custom_unit.id_
            )
        except Exception as error:
            self.record_failure(error)
            return

        if retrieved_custom_unit.id_ != custom_unit.id_:
            self.record_failure('Retrieved ID does not match')
            return

        new_name = 'CryptoJuice'

        try:
            updated_custom_unit = custom_unit.update(name=new_name)
        except Exception as error:
            self.record_failure(error)
            return

        if updated_custom_unit.name != new_name:
            self.record_failure('Name failed to update')
            return

        self.record_success()
        return



"""
Amatino API Python Bindings
PyPI Setup Module
Author: hugh@amatino.io
"""
from setuptools import setup, find_packages
from os import path
from codecs import open

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as readme_file:
    LONG_DESCRIPTION = readme_file.read()

setup(
    name='amatino',
    version='0.0.11',
    description='Bindings for the Amatino API, an accounting & financial data engine',
    long_description=LONG_DESCRIPTION,
    url='https://amatino.io',
    author='Amatino',
    author_email='hugh@amatino.io',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Operating System :: OS Independent',
        'Topic :: Office/Business :: Financial :: Accounting',
        'Topic :: Software Development :: Libraries'
    ],
    keywords='amatino api bindings accounting double-entry accounts library',
    packages=find_packages(exclude=('tests', 'tests.*')),
    long_description_content_type="text/markdown",
    python_requires='>=3',
    install_requires=['typing'],
    project_urls={
        'Twitter': 'https://twitter.com/amatinoapi',
        'Github Repository': 'https://github.com/amatino-code/amatino-python',
        'Subscribe': 'https://amatino.io/subscribe',
        'Getting Started Guide': 'https://amatino.io/articles/getting-started',
        'Discuss': 'https://amatino.io/discussion',
        'Development Newsletter': 'https://amatino.io/newsletter'
    }
)
